"""
These tests use local files in the desi_test_data/ directory that are reduced
from these real data files:

* https://data.desi.lbl.gov/public/edr/spectro/redux/fuji/healpix/sv3/dark/260/26065/coadd-sv3-dark-26065.fits
* https://data.desi.lbl.gov/public/edr/spectro/redux/fuji/healpix/sv3/dark/260/26065/spectra-sv3-dark-26065.fits
* https://data.desi.lbl.gov/public/edr/spectro/redux/fuji/tiles/cumulative/169/20210419/coadd-5-169-thru20210419.fits
* https://data.desi.lbl.gov/public/edr/spectro/redux/fuji/tiles/cumulative/169/20210419/spectra-5-169-thru20210419.fits

Tests that use the real data files are prohibitively slow because of their
size, >~ 1 GB. The reduced files are generated by generate_desi_test_data.py.
"""
from astropy.io import fits
from astropy.utils.exceptions import AstropyUserWarning
import pytest

# Package
from ..desi import (coadd_loader, spectra_loader)
from ....spectra import SpectrumList


has_importlib = True
try:
    from importlib.resources import files
except ImportError:
    from pkg_resources import resource_filename as files
    has_importlib = False


@pytest.fixture(scope="function")
def local_filename(request):
    if has_importlib:
        return files('specutils.io.default_loaders.tests') / 'desi_test_data' / request.param
    else:
        return files('specutils.io.default_loaders.tests', f'desi_test_data/{request.param}')


@pytest.mark.parametrize('local_filename, loader', [('coadd-sv3-dark-26065.fits', coadd_loader),
                                                    ('spectra-sv3-dark-26065.fits', spectra_loader),
                                                    ('coadd-5-169-thru20210419.fits', coadd_loader),
                                                    ('spectra-5-169-thru20210419.fits', spectra_loader)], indirect=['local_filename'])
def test_loader(local_filename, loader):
    """Test the 'bare' loader function in desi.py.
    """
    spectrum = loader(local_filename)  # noqa
    assert spectrum[0].meta['band'] == 'b'
    assert spectrum[1].meta['band'] == 'r'
    assert spectrum[2].meta['band'] == 'z'


@pytest.mark.parametrize('local_filename', ['coadd-sv3-dark-26065.fits',
                                            'spectra-sv3-dark-26065.fits',
                                            'coadd-5-169-thru20210419.fits',
                                            'spectra-5-169-thru20210419.fits'], indirect=['local_filename'])
def test_SpectrumList_reader(local_filename):
    """Test the class method. This also triggers the identify functions.
    """
    spectrum = SpectrumList.read(local_filename)  # noqa
    assert spectrum[0].meta['band'] == 'b'
    assert spectrum[1].meta['band'] == 'r'
    assert spectrum[2].meta['band'] == 'z'


@pytest.mark.parametrize('local_filename, fmt', [('coadd-sv3-dark-26065.fits', 'DESI coadd'),
                                                 ('spectra-sv3-dark-26065.fits', 'DESI spectra'),
                                                 ('coadd-5-169-thru20210419.fits', 'DESI coadd'),
                                                 ('spectra-5-169-thru20210419.fits', 'DESI spectra')], indirect=['local_filename'])
def test_SpectrumList_reader_fileobj(local_filename, fmt):
    """Test reading an open file-like object.
    """
    with open(local_filename, 'rb') as fileobj:
        spectrum = SpectrumList.read(fileobj, format=fmt)  # noqa
    assert spectrum[0].meta['band'] == 'b'
    assert spectrum[1].meta['band'] == 'r'
    assert spectrum[2].meta['band'] == 'z'


@pytest.mark.parametrize('local_filename, fmt', [('coadd-sv3-dark-26065.fits', 'DESI coadd'),
                                                 ('spectra-sv3-dark-26065.fits', 'DESI spectra'),
                                                 ('coadd-5-169-thru20210419.fits', 'DESI coadd'),
                                                 ('spectra-5-169-thru20210419.fits', 'DESI spectra')], indirect=['local_filename'])
def test_SpectrumList_reader_hdulist(local_filename, fmt):
    """Test reading a HDUList.
    """
    with fits.open(local_filename, mode='readonly') as hdulist:
        spectrum = SpectrumList.read(hdulist, format=fmt)  # noqa
    assert spectrum[0].meta['band'] == 'b'
    assert spectrum[1].meta['band'] == 'r'
    assert spectrum[2].meta['band'] == 'z'


@pytest.mark.parametrize('local_filename, fmt', [('coadd-sv3-dark-26065.fits', 'DESI coadd')], indirect=['local_filename'])
def test_trigger_missing_extname(local_filename, fmt):
    """Trigger a warning about a HDU not having an EXTNAME.
    """
    with fits.open(local_filename, mode='readonly') as hdulist:
        del hdulist[2].header['EXTNAME']
        with pytest.warns(AstropyUserWarning) as w:
            spectrum = SpectrumList.read(hdulist, format=fmt)  # noqa
        assert len(w) > 0
        assert str(w[0].message) == "HDU 2 has no EXTNAME and will not be read!"


@pytest.mark.parametrize('local_filename, fmt', [('coadd-sv3-dark-26065.fits', 'DESI coadd')], indirect=['local_filename'])
def test_trigger_unexpected_extname(local_filename, fmt):
    """Trigger a warning about a HDU having an unexpected EXTNAME.
    """
    with fits.open(local_filename, mode='readonly') as hdulist:
        hdulist[2].header['EXTNAME'] = ('EXT_FIBERMAP', 'unexpected extname')
        with pytest.warns(AstropyUserWarning) as w:
            spectrum = SpectrumList.read(hdulist, format=fmt)  # noqa
        assert len(w) > 0
        assert str(w[0].message) == "HDU 2, EXTNAME='EXT_FIBERMAP' is unexpected and will not be read!"



from specutils import SpectrumList
from specutils.io.registers import data_loader

from .dc_common import FITS_FILE_EXTS, SINGLE_SPLIT_LABEL
from ..parsing_utils import read_fileobj_or_hdulist

WIGGLEZ_CONFIG = {
    "hdus": None,
    "wcs": None,
    "units": None,
    "all_standard_units": True,
    "all_keywords": True,
    "valid_wcs": True,
}


def identify_wigglez(origin, *args, **kwargs):
    """
    Identify if the current file is a WiggleZ file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        if '2018MNRAS.474.4151D' in hdulist[0].header.get("REFCODE"):
            return True
    return False


@data_loader(
    label="WiggleZ", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_wigglez, priority=10,
)
def wigglez_loader(fname):
    spectra = SpectrumList.read(
        fname, format=SINGLE_SPLIT_LABEL, **WIGGLEZ_CONFIG
    )
    return spectra



from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.table import Table
from astropy.units import Quantity

from ...spectra import Spectrum1D
from ..registers import data_loader
from ..parsing_utils import read_fileobj_or_hdulist


def identify_muscles_sed(origin, *args, **kwargs):
    # check if file can be opened with this reader
    # args[0] = filename
    # fits.open(args[0]) = hdulist
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        # Test if fits has extension of type BinTable and check against
        # known keys of already defined specific formats
        return (len(hdulist) > 1 and
                isinstance(hdulist[1], fits.BinTableHDU) and
                hdulist[0].header.get('TELESCOP') == 'MULTI' and
                hdulist[0].header.get('HLSPACRN') == 'MUSCLES' and
                hdulist[0].header.get('PROPOSID') == 13650)


@data_loader(
    label="MUSCLES SED", identifier=identify_muscles_sed, dtype=Spectrum1D,
    extensions=['fits'], priority=10,
)
def muscles_sed(file_obj, **kwargs):
    """
    Load spectrum from a MUSCLES Treasury Survey panchromatic SED FITS file.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    data: Spectrum1D
        The spectrum that is represented by the data in this table.
    """
    # name is not used; what was it for?
    # name = os.path.basename(file_name.rstrip(os.sep)).rsplit('.', 1)[0]

    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header

        tab = Table.read(hdulist[1])

    meta = {'header': header}
    uncertainty = StdDevUncertainty(tab["ERROR"])
    data = Quantity(tab["FLUX"])
    wavelength = Quantity(tab["WAVELENGTH"])

    return Spectrum1D(flux=data, spectral_axis=wavelength,
                      uncertainty=uncertainty, meta=meta)



import numpy as np
import pytest

from astropy.io import fits
from astropy.units import Unit, Angstrom

from specutils import Spectrum1D, SpectrumList


def generate_apogee_hdu(observatory="APO", with_wl=True, datasum="0"):
    wl = (10**(4.179 + 6e-6 * np.arange(8575))).reshape((1, -1))
    flux = np.zeros_like(wl)
    ivar = np.zeros_like(wl)
    pixel_flags = np.zeros_like(wl)
    continuum = np.zeros_like(wl)
    nmf_rectified_model_flux = np.zeros_like(wl)

    columns = [
        fits.Column(name="spectrum_pk_id", array=[159783564], format="K"),
        fits.Column(name="release", array=[b"sdss5"], format="5A"),
        fits.Column(name="filetype", array=[b"apStar"], format="6A"),
        fits.Column(name="v_astra", array=[b"0.5.0"], format="5A"),
        fits.Column(name="healpix", array=[3], format="J"),
        fits.Column(name="sdss_id", array=[42], format="K"),
        fits.Column(name="apred", array=[b"1.2"], format="3A"),
        fits.Column(name="obj", array=[b"2M19534321+6705175"], format="18A"),
        fits.Column(name="telescope", array=[b"apo25m"], format="6A"),
        fits.Column(name="min_mjd", array=[59804], format="J"),
        fits.Column(name="max_mjd", array=[59866], format="J"),
        fits.Column(name="n_entries", array=[-1], format="J"),
        fits.Column(name="n_visits", array=[5], format="J"),
        fits.Column(name="n_good_visits", array=[5], format="J"),
        fits.Column(name="n_good_rvs", array=[5], format="J"),
        fits.Column(name="snr", array=[46.56802], format="E"),
        fits.Column(name="mean_fiber", array=[256.0], format="E"),
        fits.Column(name="std_fiber", array=[0.0], format="E"),
        fits.Column(name="spectrum_flags", array=[1048576], format="J"),
        fits.Column(name="v_rad", array=[-56.7284381], format="E"),
        fits.Column(name="e_v_rad", array=[5.35407624], format="E"),
        fits.Column(name="std_v_rad", array=[10.79173857], format="E"),
        fits.Column(name="median_e_v_rad", array=[16.19418386], format="E"),
        fits.Column(name="doppler_teff", array=[7169.0107], format="E"),
        fits.Column(name="doppler_e_teff", array=[9.405238], format="E"),
        fits.Column(name="doppler_logg", array=[2.981389], format="E"),
        fits.Column(name="doppler_e_logg", array=[0.01916536], format="E"),
        fits.Column(name="doppler_fe_h", array=[-1.20532212], format="E"),
        fits.Column(name="doppler_e_fe_h", array=[0.0093738], format="E"),
        fits.Column(name="doppler_rchi2", array=[1.1424173], format="E"),
        fits.Column(name="doppler_flags", array=[0], format="J"),
        fits.Column(name="xcorr_v_rad", array=[np.nan], format="E"),
        fits.Column(name="xcorr_v_rel", array=[np.nan], format="E"),
        fits.Column(name="xcorr_e_v_rel", array=[np.nan], format="E"),
        fits.Column(name="ccfwhm", array=[np.nan], format="E"),
        fits.Column(name="autofwhm", array=[np.nan], format="E"),
        fits.Column(name="n_components", array=[1], format="J"),
    ]
    if with_wl:
        columns.append(
            fits.Column(name="wavelength",
                        array=wl,
                        format="8575E",
                        dim="(8575)"))
    columns += [
        fits.Column(name="flux", array=flux, format="8575E", dim="(8575)"),
        fits.Column(name="ivar", array=ivar, format="8575E", dim="(8575)"),
        fits.Column(name="pixel_flags",
                    array=pixel_flags,
                    format="8575E",
                    dim="(8575)"),
        fits.Column(name="continuum",
                    array=continuum,
                    format="8575E",
                    dim="(8575)"),
        fits.Column(
            name="nmf_rectified_model_flux",
            array=nmf_rectified_model_flux,
            format="8575E",
            dim="(8575)",
        ),
        fits.Column(name="nmf_rchi2", array=[2.3391197], format="E"),
        fits.Column(name="nmf_flags", array=[0], format="J"),
    ]
    header = fits.Header(cards=[
        ("EXTNAME", f"APOGEE/{observatory}", ""),
        ("OBSRVTRY", observatory, None),
        ("INSTRMNT", "APOGEE", None),
        ("CRVAL", 4.179, None),
        ("CDELT", 6e-6, None),
        ("CTYPE", "LOG-LINEAR", None),
        ("CUNIT", "Angstrom (Vacuum)"),
        ("CRPIX", 1, None),
        ("DC-FAG", 1, None),
        ("NPIXELS", 8575, None),
        ("DATASUM", datasum, "data unit checksum updated 2023-11-13T03:21:47"),
    ])

    return fits.BinTableHDU.from_columns(columns, header=header)


def generate_boss_hdu(observatory="APO", with_wl=True, datasum="0"):
    wl = (10**(3.5523 + 1e-4 * np.arange(4648))).reshape((1, -1))
    flux = ivar = continuum = pixel_flags = nmf_rectified_model_flux = np.zeros_like(
        wl)
    columns = [
        fits.Column(name="spectrum_pk_id", array=[0], format="K"),
        fits.Column(name="release", array=["sdss5"], format="5A"),
        fits.Column(name="filetype", array=["specFull"], format="7A"),
        fits.Column(name="v_astra", array=["0.5.0"], format="5A"),
        fits.Column(name="healpix", array=[34], format="J"),
        fits.Column(name="sdss_id", array=[42], format="K"),
        fits.Column(name="run2d", array=["6_1_2"], format="6A"),
        fits.Column(name="telescope", array=["apo25m"], format="6A"),
        fits.Column(name="min_mjd", array=[54], format="J"),
        fits.Column(name="max_mjd", array=[488], format="J"),
        fits.Column(name="n_visits", array=[1], format="J"),
        fits.Column(name="n_good_visits", array=[1], format="J"),
        fits.Column(name="n_good_rvs", array=[1], format="J"),
        fits.Column(name="v_rad", array=[0], format="E"),
        fits.Column(name="e_v_rad", array=[1], format="E"),
        fits.Column(name="std_v_rad", array=[1], format="E"),
        fits.Column(name="median_e_v_rad", array=[3], format="E"),
        fits.Column(name="xcsao_teff", array=[5000], format="E"),
        fits.Column(name="xcsao_e_teff", array=[10], format="E"),
        fits.Column(name="xcsao_logg", array=[4], format="E"),
        fits.Column(name="xcsao_e_logg", array=[3], format="E"),
        fits.Column(name="xcsao_fe_h", array=[0], format="E"),
        fits.Column(name="xcsao_e_fe_h", array=[5], format="E"),
        fits.Column(name="xcsao_meanrxc", array=[0], format="E"),
        fits.Column(name="snr", array=[50], format="E"),
        fits.Column(name="gri_gaia_transform_flags", array=[1], format="J"),
        fits.Column(name="zwarning_flags", array=[0], format="J"),
    ]
    if with_wl:
        columns.append(
            fits.Column(name="wavelength",
                        array=wl,
                        format="4648E",
                        dim="(4648)"))
    columns += [
        fits.Column(name="flux", array=flux, format="4648E", dim="(4648)"),
        fits.Column(name="ivar", array=ivar, format="4648E", dim="(4648)"),
        fits.Column(name="pixel_flags",
                    array=pixel_flags,
                    format="4648E",
                    dim="(4648)"),
        fits.Column(name="continuum",
                    array=continuum,
                    format="4648E",
                    dim="(4648)"),
        fits.Column(
            name="nmf_rectified_model_flux",
            array=nmf_rectified_model_flux,
            format="4648E",
            dim="(4648)",
        ),
        fits.Column(name="nmf_rchi2", array=[5], format="E"),
        fits.Column(name="nmf_flags", array=[0], format="J"),
    ]
    header = fits.Header(cards=[
        ("EXTNAME", f"BOSS/{observatory}", ""),
        ("OBSRVTRY", observatory, None),
        ("INSTRMNT", "BOSS", None),
        ("CRVAL", 3.5523, None),
        ("CDELT", 1e-4, None),
        ("CTYPE", "LOG-LINEAR", None),
        ("CUNIT", "Angstrom (Vacuum)"),
        ("CRPIX", 1, None),
        ("DC-FAG", 1, None),
        ("NPIXELS", 4648, None),
        ("DATASUM", datasum, "data unit checksum updated 2023-11-13T03:21:47"),
    ])

    return fits.BinTableHDU.from_columns(columns, header=header)


def fake_primary_hdu():
    return fits.PrimaryHDU(header=fits.Header(cards=[
        ("SIMPLE", True, "conforms to FITS standard"),
        ("BITPIX", 8, "array data type"),
        ("NAXIS", 0, "number of array dimensions"),
        ("EXTEND", True, ""),
        ("", "", ""),
        ("", "Metadata", ""),
        ("", "", ""),
        ("V_ASTRA", "0.5.0", "Astra version"),
        (
            "CREATED",
            "23-11-13 10:21:47",
            "File creation time (UTC %y-%m-%d %H:%M:%S)",
        ),
        ("", "", ""),
        ("", "Identifiers", ""),
        ("", "", ""),
        ("SDSS_ID", 34, "SDSS-5 unique identifier"),
        ("APOGEEID", "", "SDSS-4 DR17 APOGEE identifier"),
        ("GAIA2_ID", 23423, "Gaia DR2 source identifier"),
        ("GAIA3_ID", 23423, "Gaia DR3 source identifier"),
        ("TIC_ID", 453, "TESS Input Catalog (v8) identifier"),
        ("HEALPIX", 7769, "HEALPix (128 side)"),
        ("", "", ""),
        ("", "Targeting Provenance", ""),
        ("", "", ""),
        ("CARTON_0", "", "Highest priority carton name"),
        ("LEAD", "tic_v8", "Lead catalog used for cross-match"),
        ("VER_ID", 25, "SDSS catalog version for targeting"),
        (
            "CAT_ID",
            27021597842679140,
            "Catalog identifier used to target the source",
        ),
        ("CAT_ID21", 4283338543, "Catalog identifier (v21; v0.0)"),
        ("CAT_ID25", 27021597842679140, "Catalog identifier (v25; v0.5)"),
        ("CAT_ID31", 63050395058384891, "Catalog identifier (v31; v1.0)"),
        ("N_ASSOC", 1, "SDSS_IDs associated with this CATALOGID"),
        ("N_NEIGH", 0, 'Sources within 3" and G_MAG < G_MAG_source + 5'),
        ("", "", ""),
        ("", "Astrometry", ""),
        ("", "", ""),
        ("RA", 198.4346, "Right ascension [deg]"),
        ("DEC", 67.110405, "Declination [deg]"),
        ("L", 39.57304281336641, "Galactic longitude [deg]"),
        ("B", 69.09476252445475, "Galactic latitude [deg]"),
        ("PLX", 6.2856047, "Parallax [mas]"),
        ("E_PLX", 1.04624034, "Error on parallax [mas]"),
        ("PMRA", -3.6440573, "Proper motion in RA [mas/yr]"),
        ("E_PMRA", 6.05078333, "Error on proper motion in RA [mas/yr]"),
        ("PMDE", -8.818821, "Proper motion in DEC [mas/yr]"),
        ("E_PMDE", 1.05103116, "Error on proper motion in DEC [mas/yr]"),
        ("V_RAD", "NaN", "Gaia radial velocity [km/s]"),
        ("E_V_RAD", "NaN", "Error on Gaia radial velocity [km/s]"),
        ("", "", ""),
        ("", "Gaia Photometry", ""),
        ("", "", ""),
        ("G_MAG", 5.212288, "Gaia DR3 mean G band magnitude [mag]"),
        ("BP_MAG", 1.417452, "Gaia DR3 mean BP band magnitude [mag]"),
        ("RP_MAG", 13.138654, "Gaia DR3 mean RP band magnitude [mag]"),
        ("", "", ""),
        ("", "2MASS Photometry", ""),
        ("", "", ""),
        ("J_MAG", 3.709, "2MASS J band magnitude [mag]"),
        ("E_J_MAG", 0.026, "Error on 2MASS J band magnitude [mag]"),
        ("H_MAG", 2.893, "2MASS H band magnitude [mag]"),
        ("E_H_MAG", 0.032, "Error on 2MASS H band magnitude [mag]"),
        ("K_MAG", 1.776, "2MASS K band magnitude [mag]"),
        ("E_K_MAG", 0.026, "Error on 2MASS K band magnitude [mag]"),
        ("PH_QUAL", "", "2MASS photometric quality flag"),
        ("BL_FLG", "", "Number of components fit per band (JHK)"),
        ("CC_FLG", "", "Contamination and confusion flag"),
        (
            "COMMENT",
            "See https://www.ipac.caltech.edu/2mass/releases/allsky/doc/sec2_2a.html",
            "",
        ),
        ("", "", ""),
        ("", "unWISE Photometry", ""),
        ("", "", ""),
        ("W1_MAG", 1.615836388268159, "W1 magnitude"),
        ("E_W1_MAG", 3.001952228308748062, "Error on W1 magnitude"),
        ("W1_FLUX", 954.997, "W1 flux [Vega nMgy]"),
        ("W1_DFLUX", 1.1017, "Error on W1 flux [Vega nMgy]"),
        ("W1_FRAC", 0.976102, "Fraction of W1 flux from this object"),
        ("W2_MAG", 12.20587252800215, "W2 magnitude [Vega]"),
        ("E_W2_MAG", 0.04519932356304733, "Error on W2 magnitude"),
        ("W2_FLUX", 868.906, "W2 flux [Vega nMgy]"),
        ("W2_DFLUX", 3.172016, "Error on W2 flux [Vega nMgy]"),
        ("W2_FRAC", 0.9741495, "Fraction of W2 flux from this object"),
        ("W1UFLAGS", 0, "unWISE flags for W1"),
        ("W2UFLAGS", 0, "unWISE flags for W2"),
        ("W1AFLAGS", 0, "Additional flags for W1"),
        ("W2AFLAGS", 0, "Additional flags for W2"),
        ("COMMENT", "See https://catalog.unwise.me/catalogs.html", ""),
        ("", "", ""),
        ("", "GLIMPSE Photometry", ""),
        ("", "", ""),
        ("MAG4_5", "", "IRAC band 4.5 micron magnitude [mag]"),
        ("D4_5M", "", "Error on IRAC band 4.5 micron magnitude [mag]"),
        ("RMS_F4_5", "", "RMS deviations from final flux [mJy]"),
        ("SQF_4_5", 0, "Source quality flag for IRAC band 4.5 micron"),
        ("MF4_5", 0, "Flux calculation method flag"),
        ("CSF", 0, "Close source flag"),
        (
            "COMMENT",
            "See https://irsa.ipac.caltech.edu/data/SPITZER/GLIMPSE/gator_docs/",
            "",
        ),
        ("", "", ""),
        ("", "Observations Summary", ""),
        ("", "", ""),
        ("N_BOSS", 1, "Number of BOSS visits"),
        ("B_MINMJD", 59804, "Minimum MJD of BOSS visits"),
        ("N_MAXMJD", 59866, "Maximum MJD of BOSS visits"),
        ("N_APOGEE", 1, "Number of APOGEE visits"),
        ("A_MINMJD", 59804, "Minimum MJD of APOGEE visits"),
        ("A_MAXMJD", 59866, "Maximum MJD of APOGEE visits"),
        ("", "", ""),
        ("", "Data Integrity", ""),
        ("", "", ""),
        (
            "CHECKSUM",
            "UKTRaHQOWHQOaHQO",
            "HDU checksum updated 2023-11-13T03:21:47",
        ),
        ("DATASUM", "0", "data unit checksum updated 2023-11-13T03:21:47"),
        ("", "", ""),
        ("", "HDU Descriptions", ""),
        ("", "", ""),
        ("COMMENT", "HDU 0: Summary information only", ""),
        ("COMMENT", "HDU 1: BOSS spectra from Apache Point Observatory", ""),
        ("COMMENT", "HDU 2: BOSS spectra from Las Campanas Observatory", ""),
        ("COMMENT", "HDU 3: APOGEE spectra from Apache Point Observatory", ""),
        ("COMMENT", "HDU 4: APOGEE spectra from Las Campanas Observatory", ""),
    ]))


def mwm_HDUList(hduflags, with_wl):
    hdulist = [fake_primary_hdu()]
    for i, flag in enumerate(hduflags):
        obs = ["APO", "LCO"]
        if i <= 1:
            hdulist.append(
                generate_boss_hdu(obs[i % 2],
                                  with_wl=with_wl,
                                  datasum=str(flag)))
        else:
            hdulist.append(
                generate_apogee_hdu(obs[i % 2],
                                    with_wl=with_wl,
                                    datasum=str(flag)))

    print(hdulist)
    return fits.HDUList(hdulist)


def apStar_HDUList(n_spectra):
    """Mock an apStar HDUList of n_spectra spectra."""
    np.random.seed(20)

    # init primary hdu header
    hdr = fits.Header()
    hdr["FOOBAR"] = "barfoo"
    hdr["V_APRED"] = "q"
    hdr["APRED"] = 1.3
    hdr["SNR"] = 40
    hdr["NVISITS"] = n_spectra

    # Init hdulist
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU(header=hdr))

    # Init the key HDU's (flux, error, bitmask)
    # names
    units = [
        "Flux (10^-17 erg/s/cm^2/Ang)", "Err (10^-17 erg/s/cm^2/Ang)", "Mask"
    ] + (["test"] * 7)
    for i in range(10):
        hdu = fits.ImageHDU(data=np.random.random((n_spectra, 10)))
        hdu.header["NAXIS"] = 2
        hdu.header["NAXIS1"] = 10
        hdu.header["NAXIS2"] = n_spectra
        hdu.header["CDELT1"] = 6e-06
        hdu.header["CRVAL1"] = 4.179
        hdu.header["BUNIT"] = units[i]
        hdu.name = f"apstar{i}"
        hdulist.append(hdu)

    return hdulist


def apVisit_HDUList():
    """Mock an apVisit HDUList"""
    np.random.seed(20)

    # init primary hdu header
    hdr = fits.Header()
    hdr["FOOBAR"] = "barfoo"
    hdr["SURVEY"] = "SDSS-V"
    hdr["MJD5"] = 99999
    hdr["DATE-OBS"] = "1970-01-01"

    # Init hdulist
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU(header=hdr))

    # Init the key HDU's (flux, error, bitmask, spectral)
    for i in range(4):
        if i == 3:
            hdu = fits.ImageHDU(data=np.array([
                np.arange(1, 11, 1),
                np.arange(11, 21, 1),
                np.arange(21, 31, 1)
            ]))
            hdu.header["BUNIT"] = "Wavelength (Ang)"

        else:
            hdu = fits.ImageHDU(data=np.random.random((3, 10)))
            hdu.header["BUNIT"] = "Flux (10^-17 erg/s/cm^2/Ang)"
        hdulist.append(hdu)

    return hdulist


def spec_HDUList(n_spectra):
    """Mock an BOSS spec HDUList of n_spectra spectra + 1 coadd."""
    np.random.seed(20)

    # init primary hdu header
    hdr = fits.Header()
    hdr["OBSERVAT"] = "APO"
    hdr["TELESCOP"] = "SDSS 2.5-M"
    hdr["FOOBAR"] = "barfoo"
    hdr["MJD5"] = 99999
    hdr["DATE-OBS"] = "1970-01-01"

    # Init hdulist
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU(header=hdr))

    # Init the key HDU's (flux, error, bitmask, spectral)
    names = ["COADD", "SPALL", "ZALL", "ZLINE"]
    for i in range(4):
        hdu = fits.BinTableHDU.from_columns([
            fits.Column(name="FLUX", format="E", array=np.random.random(10)),
            fits.Column(name="LOGLAM",
                        format="E",
                        array=np.random.random(10).sort()),
            fits.Column(name="IVAR", format="E", array=np.random.random(10)),
            fits.Column(name="AND_MASK",
                        format="E",
                        array=np.random.random(10)),
            fits.Column(name="OR_MASK", format="E",
                        array=np.random.random(10)),
        ])
        hdu.name = names[i]
        hdulist.append(hdu)
    for i in range(n_spectra):
        hdu = fits.BinTableHDU.from_columns([
            fits.Column(name="LOGLAM",
                        format="E",
                        array=np.random.random(10).sort()),
            fits.Column(name="FLUX", format="E", array=np.random.random(10)),
            fits.Column(name="IVAR", format="E", array=np.random.random(10)),
            fits.Column(name="AND_MASK",
                        format="E",
                        array=np.random.random(10)),
            fits.Column(name="OR_MASK", format="E",
                        array=np.random.random(10)),
        ])
        hdu.name = f"spectrum{i}"
        hdulist.append(hdu)

    return hdulist


# TEST MWM loaders
@pytest.mark.parametrize(
    "file_obj, hdu, with_wl, hduflags",
    [
        ("mwm-temp", None, False, [0, 0, 1, 0]),
        ("mwm-temp", 3, False, [0, 0, 1, 0]),
        ("mwm-temp", None, True, [0, 1, 1, 0]),
        ("mwm-temp", 2, True, [0, 1, 1, 0]),
    ],
)
def test_mwm_1d(file_obj, hdu, with_wl, hduflags):
    """Test mwm Spectrum1D loader"""
    tmpfile = str(file_obj) + ".fits"
    mwm_HDUList(hduflags, with_wl).writeto(tmpfile, overwrite=True)

    if hdu is None:
        data = Spectrum1D.read(tmpfile)
    else:
        data = Spectrum1D.read(tmpfile, hdu=hdu)
    assert isinstance(data, Spectrum1D)
    assert isinstance(data.meta["header"], fits.Header)
    if data.meta["instrument"].lower() == "apogee":
        length = 8575
    elif data.meta["instrument"].lower() == "boss":
        length = 4648
    else:
        raise ValueError(
            "INSTRMNT tag in test HDU header is not set properly.")
    assert len(data.spectral_axis.value) == length
    assert len(data.flux.value) == length
    assert data.spectral_axis.unit == Angstrom
    assert data.flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")


@pytest.mark.parametrize(
    "file_obj, with_wl, hduflags",
    [
        ("mwm-temp", False, [1, 1, 1, 1]),
        ("mwm-temp", True, [0, 1, 0, 1]),
    ],
)
def test_mwm_list(file_obj, with_wl, hduflags):
    """Test mwm SpectrumList loader"""
    tmpfile = str(file_obj) + ".fits"
    mwm_HDUList(hduflags, with_wl).writeto(tmpfile, overwrite=True)

    data = SpectrumList.read(tmpfile, format="SDSS-V mwm multi")
    assert isinstance(data, SpectrumList)
    for i in range(len(data)):
        assert isinstance(data[i], Spectrum1D)
        assert isinstance(data[i].meta["header"], fits.Header)
        if data[i].meta["instrument"].lower() == "apogee":
            length = 8575
        elif data[i].meta["instrument"].lower() == "boss":
            length = 4648
        else:
            raise ValueError(
                "INSTRMNT tag in test HDU header is not set properly.")
        assert len(data[i].spectral_axis.value) == length
        assert len(data[i].flux.value) == length
        assert data[i].spectral_axis.unit == Angstrom
        assert data[i].flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")


@pytest.mark.parametrize(
    "file_obj, hdu, hduflags",
    [
        ("mwm-temp", 1, [0, 1, 0, 1]),
        ("mwm-temp", 2, [1, 0, 1, 1]),
        ("mwm-temp", 3, [0, 0, 0, 1]),
        ("mwm-temp", 4, [0, 1, 1, 0]),
    ],
)
def test_mwm_1d_fail_spec(file_obj, hdu, hduflags):
    """Test mwm Spectrum1D loader fail on bad spec"""
    tmpfile = str(file_obj) + ".fits"
    mwm_HDUList(hduflags, True).writeto(tmpfile, overwrite=True)
    with pytest.raises(IndexError):
        Spectrum1D.read(tmpfile, hdu=hdu)


@pytest.mark.parametrize(
    "file_obj, with_wl",
    [
        ("mwm-temp", False),
        ("mwm-temp", True),
    ],
)
def test_mwm_1d_fail(file_obj, with_wl):
    """Test mwm Spectrum1D loader fail on empty"""
    tmpfile = str(file_obj) + ".fits"
    mwm_HDUList([0, 0, 0, 0], with_wl).writeto(tmpfile, overwrite=True)

    with pytest.raises(ValueError):
        Spectrum1D.read(tmpfile)


@pytest.mark.parametrize(
    "file_obj, with_wl",
    [
        ("mwm-temp", False),
        ("mwm-temp", True),
    ],
)
def test_mwm_list_fail(file_obj, with_wl):
    """Test mwm SpectrumList loader fail on empty"""
    tmpfile = str(file_obj) + ".fits"
    mwm_HDUList([0, 0, 0, 0], with_wl).writeto(tmpfile, overwrite=True)

    with pytest.raises(ValueError):
        SpectrumList.read(tmpfile, format="SDSS-V mwm multi")


@pytest.mark.parametrize(
    "file_obj,n_spectra",
    [
        ("spec-temp", 1),
        ("spec-temp", 5),
    ],
)
def test_spec_1d(file_obj, n_spectra):
    """Test BOSS spec loader"""
    tmpfile = str(file_obj) + ".fits"
    spec_HDUList(n_spectra).writeto(tmpfile, overwrite=True)

    idxs = [1] + list(np.arange(5, 5 + n_spectra, 1))

    for i in idxs:
        data = Spectrum1D.read(tmpfile, hdu=i)
        assert isinstance(data, Spectrum1D)
        assert len(data.flux.value) == 10
        assert data.flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")

        assert len(data.spectral_axis.value) == 10
        assert data.spectral_axis.unit == Angstrom
        assert len(data.mask) == 10

        assert data[i].meta["header"].get("foobar") == "barfoo"


@pytest.mark.parametrize(
    "file_obj,n_spectra",
    [
        ("spec-temp", 0),
        ("spec-temp", 5),
    ],
)
def test_spec_list(file_obj, n_spectra):
    """Test BOSS SpectrumList loader"""
    tmpfile = str(file_obj) + ".fits"
    spec_HDUList(n_spectra).writeto(tmpfile, overwrite=True)

    data = SpectrumList.read(tmpfile, format="SDSS-V spec multi")
    assert isinstance(data, SpectrumList)
    assert len(data) == n_spectra + 1
    for i in range(n_spectra):
        assert len(data[i].flux.value) == 10
        assert data[i].flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")
        assert len(data[i].spectral_axis.value) == 10
        assert data[i].spectral_axis.unit == Angstrom
        assert len(data[i].mask) == 10
        assert data[i].meta["header"].get("foobar") == "barfoo"


@pytest.mark.parametrize(
    "file_obj,hdu",
    [
        ("spec-temp", 2),
        ("spec-temp", 3),
        ("spec-temp", 4),
    ],
    ids=["Fail on HDU2", "Fail on HDU3", "Fail on HDU4"],
)
def test_spec_1d_fail_hdu(file_obj, hdu):
    """Test if fail on reading HDU2, HDU3, or HDU4 (non-spectra)"""
    tmpfile = str(file_obj) + ".fits"
    spec_HDUList(5).writeto(tmpfile, overwrite=True)

    with pytest.raises(ValueError):
        Spectrum1D.read(tmpfile, hdu=hdu)


@pytest.mark.parametrize(
    "file_obj,idx",
    [
        ("apStar-temp", 1),
        ("apStar-temp", 4),
    ],
)
def test_apStar_1D(file_obj, idx):
    tmpfile = str(file_obj) + ".fits"
    apStar_HDUList(6).writeto(tmpfile, overwrite=True)

    data = Spectrum1D.read(tmpfile, idx=idx)
    assert isinstance(data, Spectrum1D)
    assert len(data.flux.value) == 10
    assert data.flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")

    assert len(data.spectral_axis.value) == 10
    assert data.spectral_axis.unit == Angstrom

    assert data.meta["header"].get("foobar") == "barfoo"


@pytest.mark.parametrize(
    "file_obj,n_spectra",
    [
        ("apStar-temp", 3),
        ("apStar-temp", 6),
    ],
)
def test_apStar_list(file_obj, n_spectra):
    tmpfile = str(file_obj) + ".fits"
    apStar_HDUList(n_spectra).writeto(tmpfile, overwrite=True)

    data = SpectrumList.read(tmpfile, format="SDSS-V apStar multi")
    assert isinstance(data, SpectrumList)
    assert len(data) == n_spectra
    for i in range(len(data)):
        assert isinstance(data[i], Spectrum1D)
        assert len(data[i].flux.value) == 10
        assert data[i].flux.unit == Unit("1e-17 erg / (s cm2 Angstrom)")
        assert len(data[i].spectral_axis.value) == 10
        assert data[i].meta["header"].get("foobar") == "barfoo"


@pytest.mark.parametrize(
    "file_obj",
    [
        ("apStar-temp"),
    ],
)
def test_apStar_fail_list(file_obj):
    """Test if fail on reading 1D apStar as list"""
    tmpfile = str(file_obj) + ".fits"
    apStar_HDUList(1).writeto(tmpfile, overwrite=True)

    with pytest.raises(ValueError):
        SpectrumList.read(tmpfile, format="SDSS-V apStar multi")


@pytest.mark.parametrize(
    "file_obj",
    [
        ("apVisit-temp"),
    ],
)
def test_apVisit_1D(file_obj):
    tmpfile = str(file_obj) + ".fits"
    apVisit_HDUList().writeto(tmpfile, overwrite=True)

    data = Spectrum1D.read(tmpfile)
    assert isinstance(data, Spectrum1D)
    assert np.array_equal(data.spectral_axis.value, np.arange(1, 31, 1))
    assert len(data.flux.value) == 30
    assert data.meta["header"].get("foobar") == "barfoo"


@pytest.mark.parametrize(
    "file_obj",
    [
        ("apVisit-temp"),
    ],
)
def test_apVisit_list(file_obj):
    tmpfile = str(file_obj) + ".fits"
    apVisit_HDUList().writeto(tmpfile, overwrite=True)

    data = SpectrumList.read(tmpfile, format="SDSS-V apVisit multi")
    assert isinstance(data, SpectrumList)
    assert len(data) == 3



import os
import shutil
import urllib
import warnings

import pytest
import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.io.fits.verify import VerifyWarning
from astropy.table import Table
from astropy.units import UnitsWarning
from astropy.wcs import FITSFixedWarning, WCS
from astropy.io.registry import IORegistryError
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty, InverseVariance, VarianceUncertainty
from astropy.tests.helper import quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning

from numpy.testing import assert_allclose

from .conftest import remote_access
from .. import Spectrum1D, SpectrumCollection, SpectrumList
from ..io import get_loaders_by_extension
from ..io.default_loaders import subaru_pfs_spec
from ..io.default_loaders.sdss import _sdss_wcs_to_log_wcs

# NOTE: Python can be built without bz2 or lzma.
try:
    import bz2  # noqa
except ImportError:
    HAS_BZ2 = False
else:
    HAS_BZ2 = True

try:
    import lzma  # noqa
except ImportError:
    HAS_LZMA = False
else:
    HAS_LZMA = True

EBOSS_SPECTRUM_URL = 'https://data.sdss.org/sas/dr16/eboss/spectro/redux/v5_13_0/spectra/lite/4055/spec-4055-55359-0596.fits'


def test_get_loaders_by_extension():
    loader_labels = get_loaders_by_extension('fits')

    assert len(loader_labels) > 0
    assert isinstance(loader_labels[0], str)


@remote_access([{'id': '1481190', 'filename': 'L5g_0355+11_Cruz09.fits'}])
def test_spectrum1d_GMOSfits(remote_data_path):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', (VerifyWarning, UnitsWarning))
        optical_spec_2 = Spectrum1D.read(remote_data_path, format='wcs1d-fits')

    assert len(optical_spec_2.data) == 3020


@remote_access([{'id': '1481190', 'filename': 'L5g_0355+11_Cruz09.fits'}])
def test_spectrumlist_GMOSfits(remote_data_path):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', (VerifyWarning, UnitsWarning))
        spectrum_list = SpectrumList.read(remote_data_path,
                                          format='wcs1d-fits')

    assert len(spectrum_list) == 1

    spec = spectrum_list[0]
    assert len(spec.data) == 3020


@remote_access([{'id': '1481190', 'filename': 'L5g_0355+11_Cruz09.fits'}])
def test_specific_spec_axis_unit(remote_data_path):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', (VerifyWarning, UnitsWarning))
        optical_spec = Spectrum1D.read(remote_data_path,
                                       spectral_axis_unit="Angstrom",
                                       format='wcs1d-fits')

    assert optical_spec.spectral_axis.unit == "Angstrom"


@remote_access([{'id': '2656720', 'filename': '_v1410ori_20181204_261_Forrest%20Sims.fit'}])
def test_ctypye_not_compliant(remote_data_path):
    optical_spec = Spectrum1D.read(remote_data_path,  # noqa
                                   spectral_axis_unit="Angstrom",
                                   format='wcs1d-fits')


def test_generic_ecsv_reader(tmp_path):
    # Create a small data set
    wave = np.arange(1, 1.1, 0.01)*u.AA
    flux = np.ones(len(wave))*1.e-14*u.Jy
    uncertainty = 0.01*flux
    table = Table([wave, flux, uncertainty], names=["wave", "flux", "uncertainty"])
    tmpfile = str(tmp_path / '_tst.ecsv')
    table.write(tmpfile, format='ascii.ecsv')

    # Read it in and check against the original
    spectrum = Spectrum1D.read(tmpfile, format='ECSV')
    assert spectrum.spectral_axis.unit == table['wave'].unit
    assert spectrum.flux.unit == table['flux'].unit
    assert spectrum.uncertainty.unit == table['uncertainty'].unit
    assert spectrum.spectral_axis.unit == table['wave'].unit
    assert np.all(spectrum.spectral_axis == table['wave'])
    assert np.all(spectrum.flux == table['flux'])
    assert np.all(spectrum.uncertainty.array == table['uncertainty'])


@remote_access([{'id': '1481119', 'filename': 'COS_FUV.fits'},
                {'id': '1481181', 'filename': 'COS_NUV.fits'}])
def test_hst_cos(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0

    # HDUList case
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="HST/COS")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0


@remote_access([{'id': '1481192', 'filename': 'STIS_FUV.fits'},
                {'id': '1481185', 'filename': 'STIS_NUV.fits'},
                {'id': '1481183', 'filename': 'STIS_CCD.fits'}])
def test_hst_stis(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0

    # HDUList case
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="HST/STIS")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0


@pytest.mark.remote_data
def test_manga_cube():
    url = 'https://dr17.sdss.org/sas/dr17/manga/spectro/redux/v3_1_1/8485/stack/manga-8485-1901-LOGCUBE.fits.gz'
    spec = Spectrum1D.read(url, format='MaNGA cube')

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.meta['header']['INSTRUME'] == 'MaNGA'
    assert spec.shape == (34, 34, 4563)


@pytest.mark.remote_data
def test_manga_rss():
    url = 'https://dr17.sdss.org/sas/dr17/manga/spectro/redux/v3_1_1/8485/stack/manga-8485-1901-LOGRSS.fits.gz'
    spec = Spectrum1D.read(url, format='MaNGA rss')

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.meta['header']['INSTRUME'] == 'MaNGA'
    assert spec.shape == (171, 4563)


@pytest.mark.remote_data
def test_sdss_spec(tmp_path):
    sp_pattern = 'spec-4055-55359-0596.fits'
    with urllib.request.urlopen(EBOSS_SPECTRUM_URL) as response:
        # Read from open file object
        spec = Spectrum1D.read(response, format="SDSS-III/IV spec")
        assert isinstance(spec, Spectrum1D)
        assert spec.flux.size > 0

        # TODO: make this test pass
        with pytest.raises(ValueError):
            spec.write(tmp_path / '_tst.fits')

    file_path = str(tmp_path / sp_pattern)
    with urllib.request.urlopen(EBOSS_SPECTRUM_URL) as response:
        with open(file_path, 'wb') as tmp_file:
            shutil.copyfileobj(response, tmp_file)

            # Read from local disk via filename
            spec = Spectrum1D.read(tmp_file.name)
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0

            # Read from HDUList object
            with fits.open(tmp_file.name) as hdulist:
                spec = Spectrum1D.read(hdulist)
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0

            # Read from file handle
            with open(tmp_file.name, mode='rb') as fileio:
                spec = Spectrum1D.read(fileio)
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0


@pytest.mark.remote_data
def test_sdss_spspec(tmp_path):
    sp_pattern = 'spSpec-51957-0273-016.fit'
    url = f'http://das.sdss.org/spectro/1d_26/0273/1d/{sp_pattern}'
    with urllib.request.urlopen(url) as response:
        # Read from open file object
        spec = Spectrum1D.read(response, format="SDSS-I/II spSpec")
        assert isinstance(spec, Spectrum1D)
        assert spec.flux.size > 0
        assert (spec.wavelength[0] / u.AA).value == pytest.approx(3799.268615)

    file_path = str(tmp_path / sp_pattern)
    with urllib.request.urlopen(url) as response:
        with open(file_path, 'wb') as tmp_file:
            shutil.copyfileobj(response, tmp_file)

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                spec = Spectrum1D.read(tmp_file.name)

            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0

            # Read from HDUList object
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                with fits.open(tmp_file.name) as hdulist:
                    spec = Spectrum1D.read(hdulist)
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0

            # Read from file handle
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                with open(tmp_file.name, mode='rb') as fileio:
                    spec = Spectrum1D.read(fileio)
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.size > 0


@pytest.mark.remote_data
def test_sdss_spec_stream():
    """Test direct read and recognition of SDSS-III/IV spec from remote URL,
    i.e. do not rely on filename pattern.
    """
    spec = Spectrum1D.read(EBOSS_SPECTRUM_URL)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.remote_data
def test_sdss_spspec_stream():
    """Test direct read and recognition of SDSS-I/II spSpec from remote URL,
    i.e. do not rely on filename pattern.
    """
    sdss_url = 'http://das.sdss.org/spectro/1d_26/0273/1d/spSpec-51957-0273-016.fit'
    spec = Spectrum1D.read(sdss_url)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.skipif('sys.platform.startswith("win")',
                    reason='Uncertain availability of compression utilities')
@pytest.mark.remote_data
@pytest.mark.parametrize('compress', ['gzip', 'bzip2', 'xz'])
def test_sdss_compressed(compress, tmp_path):
    """Test automatic recognition of supported compression formats.
    """
    ext = {'gzip': '.gz', 'bzip2': '.bz2', 'xz': '.xz'}
    if compress == 'bzip2' and not HAS_BZ2:
        pytest.xfail("Python installation has no bzip2 support")
    if compress == 'xz' and not HAS_LZMA:
        pytest.xfail("Python installation has no lzma support")

    # Deliberately not using standard filename pattern to test header info.
    tmp_filename = tmp_path / 'SDSS-I.fits'
    with urllib.request.urlopen('http://das.sdss.org/spectro/1d_26/0273/1d/spSpec-51957-0273-016.fit') as response:
        with open(tmp_filename, 'wb') as tmp_file:
            shutil.copyfileobj(response, tmp_file)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FITSFixedWarning)
            os.system(f'{compress} {tmp_file.name}')
            spec = Spectrum1D.read(tmp_file.name + ext[compress])

        assert isinstance(spec, Spectrum1D)
        assert spec.flux.size > 0
        assert spec.uncertainty.array.min() >= 0.0

        # Try again without compression suffix:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FITSFixedWarning)
            os.system(f'mv {tmp_file.name}{ext[compress]} {tmp_file.name}')
            spec = Spectrum1D.read(tmp_file.name)

        assert isinstance(spec, Spectrum1D)
        assert spec.flux.size > 0
        assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.remote_data
def test_sdss_spplate(tmp_path):
    """Test loading of multi-object spectrum from SDSS `spPlate` format FITS file.
    """
    with urllib.request.urlopen('http://das.sdss.org/spectro/1d_26/0273/1d/spSpec-51957-0273-016.fit') as response:
        # Read reference spectrum from open file object
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FITSFixedWarning)
            spec = Spectrum1D.read(response, format="SDSS-I/II spSpec")
        assert isinstance(spec, Spectrum1D)
        assert spec.flux.size > 0
        specid = spec.meta['header']['FIBERID']

    with urllib.request.urlopen('https://data.sdss.org/sas/dr8/sdss/spectro/redux/26/0273/spPlate-0273-51957.fits') as response:
        # Read "plate" spectrum with 2D flux array from open file object
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FITSFixedWarning)
            plate = Spectrum1D.read(response, format="SDSS spPlate")
        assert isinstance(plate, Spectrum1D)
        assert plate.flux.ndim == 2
        assert plate.flux.shape[0] == 640
        assert quantity_allclose(spec.spectral_axis, plate.spectral_axis)
        assert quantity_allclose(spec.flux, plate.flux[specid-1])

    file_path = str(tmp_path / 'spPlate-0273-51957.fits')
    with urllib.request.urlopen('https://data.sdss.org/sas/dr8/sdss/spectro/redux/26/0273/spPlate-0273-51957.fits') as response:
        with open(file_path , 'wb') as tmp_file:
            shutil.copyfileobj(response, tmp_file)

            # Read from local disk via file signature
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                plate = Spectrum1D.read(tmp_file.name, limit=32)

            assert isinstance(plate, Spectrum1D)
            assert plate.flux.ndim == 2
            assert plate.flux.shape[0] == 32
            assert quantity_allclose(spec.spectral_axis, plate.spectral_axis)
            assert quantity_allclose(spec.flux, plate.flux[specid-1])

            # Read from HDUList object
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                with fits.open(tmp_file.name) as hdulist:
                    plate = Spectrum1D.read(hdulist, limit=32)
            assert plate.flux.shape[0] == 32
            assert quantity_allclose(spec.spectral_axis, plate.spectral_axis)
            assert quantity_allclose(spec.flux, plate.flux[specid-1])

            # Read from file handle
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                with open(tmp_file.name, mode='rb') as fileio:
                    plate = Spectrum1D.read(fileio, limit=32)
            assert plate.flux.shape[0] == 32
            assert quantity_allclose(spec.spectral_axis, plate.spectral_axis)
            assert quantity_allclose(spec.flux, plate.flux[specid-1])


@pytest.mark.parametrize("name", ['file.fit', 'file.fits', 'file.dat'])
def test_no_reader_matches(name, tmp_path):
    """If no reader matches a file, check that the correct error is raised.
    This test serves a second purpose: A badly written identifier
    function might raise an error as supposed to returning False when
    it cannot identify a file.  The fact that this test passes means
    that at the very least all identifier functions that have been
    tried for that file ending did not fail with an error.
    """
    filename = str(tmp_path / name)
    with open(filename, 'w') as fp:
        fp.write('asdfadasdadvzxcv')

    with pytest.raises(IORegistryError):
        Spectrum1D.read(filename)


@pytest.mark.filterwarnings('ignore:linear Solution')
@pytest.mark.filterwarnings('ignore:Flux unit was not provided')
@remote_access([{'id': '3359174', 'filename': 'linear_fits_solution.fits'}])
def test_iraf_linear(remote_data_path):

    spectrum_1d = Spectrum1D.read(remote_data_path, format='iraf')

    assert isinstance(spectrum_1d, Spectrum1D)
    assert quantity_allclose(spectrum_1d.wavelength[0],
                             u.Quantity(3514.56625402, unit='Angstrom'))
    assert quantity_allclose(spectrum_1d.wavelength[100],
                             u.Quantity(3514.56625402, unit='Angstrom') +
                             u.Quantity(0.653432383823 * 100, unit='Angstrom'))


@pytest.mark.filterwarnings('ignore:linear Solution')
@pytest.mark.filterwarnings('ignore:non-ASCII characters are present in the FITS file header')
@remote_access([{'id': '3359180', 'filename': 'log-linear_fits_solution.fits'}])
def test_iraf_log_linear(remote_data_path):

    with pytest.raises(NotImplementedError):
        assert Spectrum1D.read(remote_data_path, format='iraf')


@pytest.mark.filterwarnings('ignore:Flux unit was not provided')
@pytest.mark.filterwarnings('ignore:Read spectral axis of shape')
@remote_access([{'id': '3359190', 'filename': 'non-linear_fits_solution_cheb.fits'}])
def test_iraf_non_linear_chebyshev(remote_data_path):
    chebyshev_model = models.Chebyshev1D(degree=2, domain=[1616, 3259])
    chebyshev_model.c0.value = 5115.64008186
    chebyshev_model.c1.value = 535.515983712
    chebyshev_model.c2.value = -0.779265625182

    wavelength_axis = chebyshev_model(range(1, 4097)) * u.angstrom

    spectrum_1d = Spectrum1D.read(remote_data_path, format='iraf')

    assert isinstance(spectrum_1d, Spectrum1D)
    assert_allclose(wavelength_axis, spectrum_1d.wavelength)

    # Read from HDUList
    with fits.open(remote_data_path) as hdulist:
        spectrum_1d = Spectrum1D.read(hdulist, format='iraf')
    assert isinstance(spectrum_1d, Spectrum1D)
    assert_allclose(wavelength_axis, spectrum_1d.wavelength)


@pytest.mark.filterwarnings('ignore:Flux unit was not provided')
@pytest.mark.filterwarnings('ignore:non-ASCII characters are present in the FITS file header')
@pytest.mark.filterwarnings('ignore:Read spectral axis of shape')
@remote_access([{'id': '3359194', 'filename': 'non-linear_fits_solution_legendre.fits'}])
def test_iraf_non_linear_legendre(remote_data_path):

    legendre_model = models.Legendre1D(degree=3, domain=[21, 4048])
    legendre_model.c0.value = 5468.67555891
    legendre_model.c1.value = 835.332144466
    legendre_model.c2.value = -6.02202094803
    legendre_model.c3.value = -1.13142953897

    wavelength_axis = legendre_model(range(1, 4143)) * u.angstrom

    spectrum_1d = Spectrum1D.read(remote_data_path, format='iraf')

    assert isinstance(spectrum_1d, Spectrum1D)
    assert_allclose(wavelength_axis, spectrum_1d.wavelength)


@pytest.mark.filterwarnings('ignore:non-ASCII characters are present in the FITS file header')
@remote_access([{'id': '3359196', 'filename': 'non-linear_fits_solution_linear-spline.fits'}])
def test_iraf_non_linear_linear_spline(remote_data_path):

    with pytest.raises(NotImplementedError):
        assert Spectrum1D.read(remote_data_path, format='iraf')


@pytest.mark.filterwarnings('ignore:non-ASCII characters are present in the FITS file header')
@remote_access([{'id': '3359200', 'filename': 'non-linear_fits_solution_cubic-spline.fits'}])
def test_iraf_non_linear_cubic_spline(remote_data_path):

    with pytest.raises(NotImplementedError):
        assert Spectrum1D.read(remote_data_path, format='iraf')


@pytest.mark.filterwarnings('ignore:Flux unit was not provided')
@pytest.mark.remote_data
def test_iraf_multispec_chebyshev():
    """Test loading of SpectrumCollection from IRAF MULTISPEC format FITS file -
    nonlinear 2D WCS with Chebyshev solution (FTYPE=1).
    """
    iraf_url = 'https://github.com/astropy/specutils/raw/legacy-specutils/specutils/io/tests/files'
    # Read reference ASCII spectrum from remote file.
    spec10 = Table.read(iraf_url + '/AAO_11.txt', format='ascii.no_header',
                        data_start=175, names=['spectral_axis', 'flux'])

    # Read full collection of 51 spectra from open FITS file object
    speccol = SpectrumCollection.read(iraf_url + '/AAO.fits')

    assert len(speccol) == 51
    # Numpy allclose does support quantities, but not with unit 'adu'!
    assert quantity_allclose(speccol[10].spectral_axis, spec10['spectral_axis'] * u.AA)
    assert quantity_allclose(speccol[10].flux, spec10['flux'] * u.adu)


@pytest.mark.filterwarnings('ignore:Flux unit was not provided')
@pytest.mark.remote_data
def test_iraf_multispec_legendre():
    """Test loading of SpectrumCollection from IRAF MULTISPEC format FITS file -
    nonlinear 2D WCS with Legendre solution (FTYPE=2).
    """
    iraf_url = 'https://github.com/astropy/specutils/raw/legacy-specutils/specutils/io/tests/files'
    # Read reference ASCII spectrum from remote file.
    spec10 = Table.read(iraf_url + '/TRES.dat', format='ascii.no_header',
                        data_start=127, names=['spectral_axis', 'flux'])

    # Read full collection of 51 spectra from remote FITS file object
    speccol = SpectrumCollection.read(iraf_url + '/TRES.fits')

    assert len(speccol) == 51
    # The reference spectrum flux is normalised/flatfielded, just check the wavelength solution
    assert_allclose(speccol[10].spectral_axis, spec10['spectral_axis'] * u.AA)


@pytest.mark.parametrize("spectral_axis",
                         ['wavelength', 'frequency', 'energy', 'wavenumber'])
def test_tabular_fits_writer(tmp_path, spectral_axis):
    wlu = {'wavelength': u.AA, 'frequency': u.GHz, 'energy': u.eV,
           'wavenumber': u.cm**-1}
    # Create a small data set
    disp = np.arange(1, 1.1, 0.01) * wlu[spectral_axis]
    flux = np.ones(len(disp)) * 1.e-14 * u.Jy
    unc = StdDevUncertainty(0.01 * flux)
    if spectral_axis not in ('wavelength', ):
        disp = np.flip(disp)

    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, uncertainty=unc)
    tmpfile = str(tmp_path / '_tst.fits')
    spectrum.write(tmpfile, format='tabular-fits')

    # Read it in and check against the original
    spec = Spectrum1D.read(tmpfile)
    assert spec.flux.unit == spectrum.flux.unit
    assert spec.spectral_axis.unit == spectrum.spectral_axis.unit
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity,
                             spectrum.uncertainty.quantity)

    # Test spectrum with different flux unit
    flux = np.random.normal(0., 1.e-9, disp.shape[0]) * u.W * u.m**-2 * u.AA**-1
    unc = StdDevUncertainty(0.1 * np.sqrt(np.abs(flux.value)) * flux.unit)
    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, uncertainty=unc)

    # Try to overwrite the file
    with pytest.raises(OSError, match=r'File .*exists'):
        spectrum.write(tmpfile, format='tabular-fits')
    spectrum.write(tmpfile, format='tabular-fits', overwrite=True)

    # Map to alternative set of units
    cmap = {spectral_axis: ('spectral_axis', 'micron'),
            'flux': ('flux', 'erg / (s cm**2 AA)'),
            'uncertainty': ('uncertainty', None)}

    # Read it back again and check against the original
    spec = Spectrum1D.read(tmpfile, format='tabular-fits', column_mapping=cmap)
    assert spec.flux.unit == u.Unit('erg / (s cm**2 AA)')
    assert spec.spectral_axis.unit == u.um
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity,
                             spectrum.uncertainty.quantity)


@pytest.mark.parametrize("ndim", range(1, 4))
@pytest.mark.parametrize("spectral_axis",
                         ['wavelength', 'frequency', 'energy', 'wavenumber'])
def test_tabular_fits_multid(tmp_path, ndim, spectral_axis):
    wlu = {'wavelength': u.AA, 'frequency': u.GHz, 'energy': u.eV,
           'wavenumber': u.cm**-1}
    # Create a small data set with ndim-D flux + uncertainty
    disp = np.arange(1, 1.1, 0.01) * wlu[spectral_axis]
    shape = (3, 2, 4)[:ndim+1] + disp.shape
    flux = np.random.normal(0., 1.e-9, shape) * u.W * u.m**-2 * u.AA**-1
    unc = StdDevUncertainty(0.01 * np.random.sample(shape))
    if spectral_axis not in ('wavelength', ):
        disp = np.flip(disp)

    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, uncertainty=unc)
    tmpfile = str(tmp_path / '_tst.fits')
    spectrum.write(tmpfile, format='tabular-fits')

    # Read it in and check against the original
    spec = Spectrum1D.read(tmpfile)
    assert spec.flux.unit == spectrum.flux.unit
    assert spec.spectral_axis.unit == spectrum.spectral_axis.unit
    assert spec.flux.shape == flux.shape
    assert spec.uncertainty.array.shape == flux.shape
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity,
                             spectrum.uncertainty.quantity)

    # Test again, using `column_mapping` to convert to different spectral axis and flux units
    cmap = {spectral_axis: ('spectral_axis', 'THz'),
            'flux': ('flux', 'erg / (s cm**2 AA)'),
            'uncertainty': ('uncertainty', None)}

    spec = Spectrum1D.read(tmpfile, format='tabular-fits', column_mapping=cmap)
    assert spec.flux.unit == u.Unit('erg / (s cm**2 AA)')
    assert spec.spectral_axis.unit == u.THz
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity,
                             spectrum.uncertainty.quantity)


@pytest.mark.parametrize("mask_type", [bool, np.uint8, np.int8, np.uint16, np.int16, '>i2'])
def test_tabular_fits_mask(tmp_path, mask_type):
    # test mask I/O with tabular fits format
    wave = np.arange(3600, 3700) * u.AA
    nwave = len(wave)

    # 1D Case
    flux = np.random.uniform(0,1,size=nwave) * u.Jy
    mask = np.zeros(flux.shape, dtype=mask_type)
    mask[0] = 1

    sp1 = Spectrum1D(spectral_axis=wave, flux=flux, mask=mask)
    assert sp1.mask.dtype == mask.dtype

    tmpfile = str(tmp_path / '_mask_tst.fits')
    sp1.write(tmpfile, format='tabular-fits', overwrite=True)

    sp2 = Spectrum1D.read(tmpfile)
    assert np.all(sp1.spectral_axis == sp2.spectral_axis)
    assert np.all(sp1.flux == sp2.flux)
    assert sp2.mask is not None
    assert np.all(sp1.mask == sp2.mask)

    # int16 is returned as FITS-native '>i2'
    if mask_type == np.int16:
        assert sp1.mask.dtype.kind == sp2.mask.dtype.kind
        assert sp1.mask.dtype.itemsize == sp2.mask.dtype.itemsize
    elif mask_type == np.int8:
        # due to https://github.com/astropy/astropy/issues/11963,
        # int8 is upcast to int16 which is returned as >i2...
        assert sp2.mask.dtype == np.dtype('>i2')
    else:
        assert sp1.mask.dtype == sp2.mask.dtype

    # 2D Case
    nspec = 3
    flux = np.random.uniform(0,1,size=(nspec,nwave)) * u.Jy
    mask = np.zeros(flux.shape, dtype=mask_type)
    mask[0,0] = 1

    sp1 = Spectrum1D(spectral_axis=wave, flux=flux, mask=mask)

    tmpfile = str(tmp_path / '_mask_tst.fits')
    sp1.write(tmpfile, format='tabular-fits', overwrite=True)

    sp2 = Spectrum1D.read(tmpfile)
    assert np.all(sp1.spectral_axis == sp2.spectral_axis)
    assert np.all(sp1.flux == sp2.flux)
    assert sp2.mask is not None
    assert np.all(sp1.mask == sp2.mask)

    # int16 is returned as FITS-native '>i2'
    if mask_type == np.int16:
        assert sp1.mask.dtype.kind == sp2.mask.dtype.kind
        assert sp1.mask.dtype.itemsize == sp2.mask.dtype.itemsize
    elif mask_type == np.int8:
        # due to https://github.com/astropy/astropy/issues/11963,
        # int8 is upcast to int16 which is returned as >i2...
        assert sp2.mask.dtype == np.dtype('>i2')
    else:
        assert sp1.mask.dtype == sp2.mask.dtype


def test_tabular_fits_maskheader(tmp_path):
    # Create a small data set + header with reserved FITS keywords
    disp = np.linspace(1, 1.2, 21) * u.AA
    flux = np.random.normal(0., 1.0e-14, disp.shape[0]) * u.Jy
    hdr = fits.header.Header({'TELESCOP': 'Leviathan', 'APERTURE': 1.8,
                              'OBSERVER': 'Parsons', 'NAXIS': 1, 'NAXIS1': 8})

    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, meta={'header': hdr})
    tmpfile = str(tmp_path / '_tst.fits')
    spectrum.write(tmpfile, format='tabular-fits')

    # Read it in and check against the original
    with fits.open(tmpfile) as hdulist:
        assert hdulist[0].header['NAXIS'] == 0
        assert hdulist[1].header['NAXIS'] == 2
        assert hdulist[1].header['NAXIS2'] == disp.shape[0]
        assert hdulist[1].header['OBSERVER'] == 'Parsons'

    # Now write with updated header information from spectrum.meta
    spectrum.meta.update({'OBSERVER': 'Rosse', 'EXPTIME': 32.1, 'NAXIS2': 12})
    spectrum.write(tmpfile, format='tabular-fits', overwrite=True,
                   update_header=True)

    with fits.open(tmpfile) as hdulist:
        assert hdulist[1].header['NAXIS2'] == disp.shape[0]
        assert hdulist[1].header['OBSERVER'] == 'Rosse'
        assert_allclose(hdulist[1].header['EXPTIME'], 3.21e1)

    # Test that unsupported types (dict) are not added to written header
    spectrum.meta['MYHEADER'] = {'OBSDATE': '1848-02-26', 'TARGET': 'M51'}
    spectrum.write(tmpfile, format='tabular-fits', overwrite=True,
                   update_header=True)

    with fits.open(tmpfile) as hdulist:
        assert 'MYHEADER' not in hdulist[0].header
        assert 'MYHEADER' not in hdulist[1].header
        assert 'OBSDATE' not in hdulist[0].header
        assert 'OBSDATE' not in hdulist[1].header


@pytest.mark.filterwarnings("ignore:The unit 'Angstrom' has been deprecated")
def test_tabular_fits_autowrite(tmp_path):
    """Test writing of Spectrum1D with automatic selection of BINTABLE format."""
    disp = np.linspace(1, 1.2, 21) * u.AA
    flux = np.random.normal(0., 1.0e-14, disp.shape[0]) * u.W / (u.m**2 * u.AA)
    hdr = fits.header.Header({'TELESCOP': 'Leviathan', 'APERTURE': 1.8,
                              'OBSERVER': 'Parsons', 'NAXIS': 1, 'NAXIS1': 8})

    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, meta={'header': hdr})
    tmpfile = str(tmp_path / '_tst.fits')
    spectrum.write(tmpfile)

    # Read it in and check against the original
    with fits.open(tmpfile) as hdulist:
        assert hdulist[0].header['NAXIS'] == 0
        assert hdulist[1].header['NAXIS'] == 2
        assert hdulist[1].header['NAXIS2'] == disp.shape[0]

    # Trigger exception for illegal HDU (primary HDU only accepts IMAGE_HDU)
    with pytest.raises(ValueError, match=r'FITS does not support BINTABLE'):
        spectrum.write(tmpfile, format='tabular-fits', overwrite=True, hdu=0)

    # Test automatic selection of wcs1d format, which will fail without suitable wcs
    with pytest.raises(ValueError, match=r'Only Spectrum1D objects with valid WCS'):
        spectrum.write(tmpfile, overwrite=True, hdu=0)

    tmpfile = str(tmp_path / '_wcs.fits')
    with pytest.raises(ValueError, match=r'Only Spectrum1D objects with valid WCS'):
        spectrum.write(tmpfile, overwrite=True)


@pytest.mark.skipif('sys.platform.startswith("win")',
                    reason='Uncertain availability of compression utilities')
@pytest.mark.parametrize('compress', ['gzip', 'bzip2', 'xz'])
def test_tabular_fits_compressed(compress, tmp_path):
    """Test automatic recognition of supported compression formats for BINTABLE.
    """
    ext = {'gzip': '.gz', 'bzip2': '.bz2', 'xz': '.xz'}
    if compress == 'bzip2' and not HAS_BZ2:
        pytest.xfail("Python installation has no bzip2 support")
    if compress == 'xz' and not HAS_LZMA:
        pytest.xfail("Python installation has no lzma support")

    # Create a small data set
    disp = np.linspace(1, 1.2, 23) * u.AA
    flux = np.random.normal(0., 1.0e-14, disp.shape[0]) * u.Jy
    unc = StdDevUncertainty(0.01 * np.abs(flux))

    spectrum = Spectrum1D(flux=flux, spectral_axis=disp, uncertainty=unc)
    tmpfile = str(tmp_path / '_tst.fits')
    spectrum.write(tmpfile, format='tabular-fits')

    # Deliberately not using standard filename pattern to test header info.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FITSFixedWarning)
        os.system(f'{compress} {tmpfile}')
        spec = Spectrum1D.read(tmpfile + ext[compress])

    assert isinstance(spec, Spectrum1D)
    assert spec.spectral_axis.shape[0] == len(disp)
    assert spec.flux.size == len(disp)
    assert spec.uncertainty.array.min() >= 0.0
    assert quantity_allclose(spec.flux, spectrum.flux)

    # Try again without compression suffix:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FITSFixedWarning)
        os.system(f'mv {tmpfile}{ext[compress]} {tmpfile}')
        spec = Spectrum1D.read(tmpfile)

    assert isinstance(spec, Spectrum1D)
    assert spec.spectral_axis.shape[0] == len(disp)
    assert spec.flux.size == len(disp)
    assert spec.uncertainty.array.min() >= 0.0
    assert quantity_allclose(spec.flux, spectrum.flux)


@pytest.mark.parametrize("spectral_axis", ['WAVE', 'FREQ', 'ENER', 'WAVN'])
@pytest.mark.parametrize("uncertainty",
                         [None, StdDevUncertainty, VarianceUncertainty, InverseVariance])
def test_wcs1d_fits_writer(tmp_path, spectral_axis, uncertainty):
    """Test write/read for Spectrum1D with WCS-constructed spectral_axis."""
    wlunits = {'WAVE': 'Angstrom', 'FREQ': 'GHz', 'ENER': 'eV', 'WAVN': 'cm**-1'}
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': spectral_axis, 'CUNIT1': wlunits[spectral_axis],
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small data set
    flux = np.arange(1, 11)**2 * 1.e-14 * u.Jy
    wlu = u.Unit(hdr['CUNIT1'])
    wl0 = hdr['CRVAL1']
    dwl = hdr['CDELT1']
    disp = np.arange(wl0, wl0 + (len(flux) - 0.5) * dwl, dwl) * wlu
    tmpfile = tmp_path / 'wcs_tst.fits'

    mask = np.array([0, 0, 0, 0, 1, 0, 0, 0, 1, 0], dtype=np.uint8)

    # ToDo: test with explicit (and different from flux) units.
    if uncertainty is None:
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask)
        assert spectrum.uncertainty is None
    else:
        unc = uncertainty(0.1 * np.sqrt(np.abs(flux.value)))
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask, uncertainty=unc)
    spectrum.write(tmpfile, hdu=0)

    # Read it in and check against the original
    spec = Spectrum1D.read(tmpfile)
    assert spec.flux.unit == spectrum.flux.unit
    assert spec.spectral_axis.unit == spectrum.spectral_axis.unit
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.spectral_axis, disp)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert np.all(spec.mask == spectrum.mask)
    if uncertainty is None:
        assert spec.uncertainty is None
    else:
        assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)

    # Read from HDUList
    with fits.open(tmpfile) as hdulist:
        spec = Spectrum1D.read(hdulist, format='wcs1d-fits')

    assert isinstance(spec, Spectrum1D)
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert np.all(spec.mask == spectrum.mask)
    if uncertainty is None:
        assert spec.uncertainty is None
    else:
        assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)


@pytest.mark.parametrize("spectral_axis", ['WAVE', 'FREQ'])
@pytest.mark.parametrize("mask_type", [None, bool, np.uint8, np.int8, np.uint16, np.int16, '>i2'])
@pytest.mark.parametrize("uncertainty", [StdDevUncertainty, InverseVariance])
def test_wcs1d_fits_masks(tmp_path, spectral_axis, mask_type, uncertainty):
    """Test write/read for Spectrum1D with WCS-constructed spectral_axis."""
    wlunits = {'WAVE': 'nm', 'FREQ': 'GHz', 'ENER': 'eV', 'WAVN': 'cm**-1'}
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': spectral_axis, 'CUNIT1': wlunits[spectral_axis],
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small data set
    flux = np.arange(1, 11)**2 * 1.e-14 * u.Jy
    wlu = u.Unit(hdr['CUNIT1'])
    wl0 = hdr['CRVAL1']
    dwl = hdr['CDELT1']
    disp = np.arange(wl0, wl0 + (len(flux) - 0.5) * dwl, dwl) * wlu
    unc = uncertainty(0.1 * np.sqrt(np.abs(flux.value)))
    tmpfile = tmp_path / 'wcs_tst.fits'

    if mask_type is None:
        mask = None
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), uncertainty=unc)
        assert spectrum.mask is None
    else:
        mask = np.array([0, 0, 1, 0, 3, 0, 0, -99, -199, 0]).astype(mask_type)
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask, uncertainty=unc)
        assert spectrum.mask.dtype == mask.dtype

    spectrum.write(tmpfile, hdu=0)

    # Read it in and check against the original
    spec = Spectrum1D.read(tmpfile)
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.spectral_axis, disp)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)
    assert np.all(spec.mask == spectrum.mask)
    # int16 is returned as FITS-native '>i2'
    if mask_type == np.int16:
        assert spec.mask.dtype.kind == spectrum.mask.dtype.kind
        assert spec.mask.dtype.itemsize == spectrum.mask.dtype.itemsize
    else:
        assert np.array(spec.mask).dtype == np.array(spectrum.mask).dtype

    # Read from HDUList
    with fits.open(tmpfile) as hdulist:
        spec = Spectrum1D.read(hdulist, format='wcs1d-fits')

    assert isinstance(spec, Spectrum1D)
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)
    assert np.all(spec.mask == spectrum.mask)
    if mask_type == np.int16:
        assert spec.mask.dtype.kind == spectrum.mask.dtype.kind
        assert spec.mask.dtype.itemsize == spectrum.mask.dtype.itemsize
    else:
        assert np.array(spec.mask).dtype == np.array(spectrum.mask).dtype


@pytest.mark.parametrize("spectral_axis",
                         ['WAVE', 'FREQ', 'ENER', 'WAVN'])
@pytest.mark.parametrize("with_mask", [False, True])
@pytest.mark.parametrize("uncertainty",
                         [None, StdDevUncertainty, VarianceUncertainty, InverseVariance])
def test_wcs1d_fits_cube(tmp_path, spectral_axis, with_mask, uncertainty):
    """Test write/read for Spectrum1D spectral cube with WCS spectral_axis."""
    wlunits = {'WAVE': 'Angstrom', 'FREQ': 'GHz', 'ENER': 'eV', 'WAVN': 'cm**-1'}
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': spectral_axis, 'CUNIT1': wlunits[spectral_axis],
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01,
           'WCSAXES': 3, 'DISPAXIS': 0,
           'CRPIX2': 38.0, 'CRPIX3': 38.0, 'CRVAL2': 205.4384, 'CRVAL3': 27.004754,
           'CTYPE2': 'RA---TAN', 'CTYPE3': 'DEC--TAN', 'CUNIT2': 'deg', 'CUNIT3': 'deg',
           'CDELT2': 3.61111097865634E-05, 'CDELT3': 3.61111097865634E-05,
           'PC1_1': 1.0, 'PC12 ': 0, 'PC1_3': 0,
           'PC2_1': 0, 'PC2_2': -1.0, 'PC2_3': 0,
           'PC3_1': 0, 'PC3_2': 0, 'PC3_3': 1.0}
    # Create a small data set
    flux = np.arange(1, 121).reshape((10, 4, 3))**2 * 1.e-14 * u.Jy
    wlu = u.Unit(hdr['CUNIT1'])
    wl0 = hdr['CRVAL1']
    dwl = hdr['CDELT1']
    disp = np.arange(wl0, wl0 + (flux.shape[2] - 0.5) * dwl, dwl) * wlu
    tmpfile = tmp_path / 'wcs_tst.fits'

    if with_mask:
        die = np.random.Generator(np.random.MT19937(23))
        mask = die.choice([0, 0, 0, 0, 0, 1], size=flux.shape).astype(np.uint16)
    else:
        mask = None

    if uncertainty is None:
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask)
        assert spectrum.uncertainty is None
        with pytest.warns(AstropyUserWarning, match='No uncertainty array found'):
            spectrum.write(tmpfile, hdu=0, uncertainty_name='STD')
    else:
        unc = uncertainty(0.1 * np.sqrt(np.abs(flux.value)))
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask, uncertainty=unc)
        spectrum.write(tmpfile, hdu=0)

    # Broken reader!
    # Read it in and check against the original
    spec = Spectrum1D.read(tmpfile, format='wcs1d-fits')
    assert spec.flux.unit == spectrum.flux.unit
    assert spec.flux.shape == spectrum.flux.shape
    assert spec.spectral_axis.unit == spectrum.spectral_axis.unit
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.spectral_axis, disp)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert np.all(spec.mask == spectrum.mask)
    if uncertainty is None:
        assert spec.uncertainty is None
    else:
        assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)

    # Read from HDUList
    with fits.open(tmpfile) as hdulist:
        w = WCS(hdulist[0].header)
        spec = Spectrum1D.read(hdulist, format='wcs1d-fits')

    assert w.naxis == 3
    assert w.axis_type_names == [spectral_axis, 'RA', 'DEC']
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.shape == spectrum.flux.shape
    assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
    assert quantity_allclose(spec.flux, spectrum.flux)
    assert np.all(spec.mask == spectrum.mask)
    if uncertainty is None:
        assert spec.uncertainty is None
    else:
        assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)


@pytest.mark.parametrize("uncertainty_rsv", ['STD', 'ERR', 'UNCERT', 'VAR', 'IVAR'])
@pytest.mark.parametrize("hdu", [None, 0, 1])
def test_wcs1d_fits_uncertainty(tmp_path, uncertainty_rsv, hdu):
    """
    Test Spectrum1D.write with custom `uncertainty` names,
    ensure it raises on illegal (reserved) names.
    """
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': 'WAVE', 'CUNIT1': 'um', 'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.001}
    # Reserved EXTNAMEs for uncertainty types
    UNCERT_REF = {'STD': StdDevUncertainty, 'ERR': StdDevUncertainty, 'UNCERT': StdDevUncertainty,
                  'VAR': VarianceUncertainty, 'IVAR': InverseVariance}
    # Alternative EXTNAMEs for uncertainty types
    UNCERT_ALT = {'std': 'StdErr', 'var': 'VARIAN', 'ivar': 'InvVar'}

    tmpfile = tmp_path / 'wcs_tst.fits'

    # Create a small data set
    flux = np.arange(1, 11)**2 * 1.e-14 * u.Jy
    mask = np.array([0, 0, 0, 0, 1, 0, 0, 0, 1, 0], dtype=np.uint16)

    # Set uncertainty to mismatched type
    uncertainty = [u for n, u in UNCERT_REF.items() if u != UNCERT_REF[uncertainty_rsv]][0]
    unc = uncertainty(0.1 * np.sqrt(np.abs(flux.value)))
    spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr), mask=mask, uncertainty=unc)

    with pytest.raises(ValueError, match=f"Illegal label for uncertainty: '{uncertainty_rsv}' "
                       f"is reserved for {UNCERT_REF[uncertainty_rsv]}, not {uncertainty}."):
        spectrum.write(tmpfile, format='wcs1d-fits', uncertainty_name=uncertainty_rsv)

    # Set permitted custom name
    uncertainty_type = spectrum.uncertainty.uncertainty_type
    uncertainty_alt = UNCERT_ALT[uncertainty_type]
    if hdu is None:
        spectrum.write(tmpfile, format='wcs1d-fits', uncertainty_name=uncertainty_alt)
        hdu = 0
    else:
        spectrum.write(tmpfile, format='wcs1d-fits', uncertainty_name=uncertainty_alt, hdu=hdu)

    # Auto-detect only works with flux in default (primary) HDU.
    if hdu == 0:
        kwargs = {'uncertainty_hdu': hdu+1}
    else:
        kwargs = {'uncertainty_hdu': hdu+1, 'format': 'wcs1d-fits'}

    # Check EXTNAME (uncertainty is in first HDU following flux spectrum)
    with fits.open(tmpfile) as hdulist:
        assert hdulist[hdu+1].name == uncertainty_alt.upper()

    # Read it in and check against the original
    with pytest.raises(ValueError, match=f"Invalid uncertainty type: '{uncertainty_alt}'; should"):
        spec = Spectrum1D.read(tmpfile, uncertainty_type=uncertainty_alt, **kwargs)
    # Need to specify type if not default
    with pytest.warns(AstropyUserWarning, match="Could not determine uncertainty type for HDU "
                      rf"'{hdu+1}' .'{uncertainty_alt.upper()}'., assuming 'StdDev'"):
        spec = Spectrum1D.read(tmpfile, **kwargs)
    if uncertainty_type != 'std':
        assert spec.uncertainty.uncertainty_type != uncertainty_type
        spec = Spectrum1D.read(tmpfile, uncertainty_type=uncertainty_type, **kwargs)

    assert spec.flux.unit == spectrum.flux.unit
    assert spec.spectral_axis.unit == spectrum.spectral_axis.unit
    assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)
    kwargs['uncertainty_hdu'] = uncertainty_alt
    spec = Spectrum1D.read(tmpfile, uncertainty_type=uncertainty_type, **kwargs)
    assert quantity_allclose(spec.uncertainty.quantity, spectrum.uncertainty.quantity)


@pytest.mark.filterwarnings('ignore:Card is too long')
@pytest.mark.parametrize("hdu", range(3))
def test_wcs1d_fits_hdus(tmp_path, hdu):
    """Test writing of Spectrum1D in WCS1D format to different IMAGE_HDUs."""
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': 'wavelength', 'CUNIT1': 'um',
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small data set
    flu = u.W / (u.m**2 * u.nm)
    flux = np.arange(1, 11)**2 * 1.e-14 * flu

    spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr))
    tmpfile = tmp_path / 'tst.fits'
    spectrum.write(tmpfile, hdu=hdu, format='wcs1d-fits')

    # Read it in and check against the original
    with fits.open(tmpfile) as hdulist:
        assert hdulist[hdu].is_image
        assert hdulist[hdu].header['NAXIS'] == 1
        assert hdulist[hdu].header['NAXIS1'] == flux.shape[0]
        assert u.Unit(hdulist[hdu].header['CUNIT1']) == u.Unit(hdr['CUNIT1'])
        assert quantity_allclose(hdulist[hdu].data * flu, flux)

    # Test again with automatic format selection by filename pattern
    tmpfile = tmp_path / 'wcs.fits'
    spectrum.write(tmpfile, hdu=hdu)
    with fits.open(tmpfile) as hdulist:
        assert hdulist[hdu].is_image
        assert quantity_allclose(hdulist[hdu].data * flu, flux)


@pytest.mark.parametrize("spectral_axis",
                         ['WAVE', 'FREQ', 'ENER', 'WAVN'])
def test_wcs1d_fits_multid(tmp_path, spectral_axis):
    """Test spectrum with WCS-1D spectral_axis and higher dimension in flux."""
    wlunits = {'WAVE': 'Angstrom', 'FREQ': 'GHz', 'ENER': 'eV', 'WAVN': 'cm**-1'}
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': spectral_axis, 'CUNIT1': wlunits[spectral_axis],
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small data set
    flux = np.arange(1, 11)**2 * 1.e-14 * u.Jy
    wlu = u.Unit(hdr['CUNIT1'])
    wl0 = hdr['CRVAL1']
    dwl = hdr['CDELT1']
    disp = np.arange(wl0, wl0 + len(flux[1:]) * dwl, dwl) * wlu

    # Construct 2D to 4D flux array, write and read (no auto-identify)
    shape = [-1, 1]
    for i in range(2, 5):
        flux = flux * np.arange(i, i+5).reshape(*shape)
        spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr))
        tmpfile = tmp_path / f'wcs_{i}d.fits'
        spectrum.write(tmpfile, format='wcs1d-fits')

        spec = Spectrum1D.read(tmpfile, format='wcs1d-fits')
        assert spec.flux.ndim == i
        assert quantity_allclose(spec.spectral_axis, disp)
        assert quantity_allclose(spec.spectral_axis, spectrum.spectral_axis)
        assert quantity_allclose(spec.flux, spectrum.flux)
        shape.append(1)

    # Test exception for NAXIS > 4
    flux = flux * np.arange(i+1, i+6).reshape(*shape)
    spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr))
    tmpfile = tmp_path / f'wcs_{i+1}d.fits'
    spectrum.write(tmpfile, format='wcs1d-fits')

    with pytest.raises(ValueError, match='input to wcs1d_fits_loader is > 4D'):
        spec = Spectrum1D.read(tmpfile, format='wcs1d-fits')


@pytest.mark.parametrize("spectral_axis", ['WAVE', 'FREQ'])
def test_wcs1d_fits_non1d(tmp_path, spectral_axis):
    """Test exception on trying to load FITS with 2D flux and irreducible WCS
    spectral_axis.
    """
    wlunits = {'WAVE': 'Angstrom', 'FREQ': 'GHz', 'ENER': 'eV', 'WAVN': 'cm**-1'}
    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': spectral_axis, 'CUNIT1': wlunits[spectral_axis],
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small 2D data set
    flux = np.arange(1, 11)**2 * np.arange(4).reshape(-1, 1) * 1.e-14 * u.Jy
    spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr))
    tmpfile = tmp_path / f'wcs_{2}d.fits'
    spectrum.write(tmpfile, format='wcs1d-fits')

    # Reopen file and update header with off-diagonal element
    with fits.open(tmpfile, mode='update') as hdulist:
        hdulist[0].header.update([('PC1_2', 0.2)])

    with pytest.raises(ValueError,
                       match='Non-zero off-diagonal matrix elements excluded from the subimage.'):
        Spectrum1D.read(tmpfile, format='wcs1d-fits')


@pytest.mark.skipif('sys.platform.startswith("win")',
                    reason='Uncertain availability of compression utilities')
@pytest.mark.parametrize('compress', ['gzip', 'bzip2', 'xz'])
def test_wcs1d_fits_compressed(compress, tmp_path):
    """Test automatic recognition of supported compression formats for IMAGE/WCS.
    """
    ext = {'gzip': '.gz', 'bzip2': '.bz2', 'xz': '.xz'}
    if compress == 'bzip2' and not HAS_BZ2:
        pytest.xfail("Python installation has no bzip2 support")
    if compress == 'xz' and not HAS_LZMA:
        pytest.xfail("Python installation has no lzma support")

    # Header dictionary for constructing WCS
    hdr = {'CTYPE1': 'wavelength', 'CUNIT1': 'Angstrom',
           'CRPIX1': 1, 'CRVAL1': 1, 'CDELT1': 0.01}
    # Create a small data set
    flux = np.arange(1, 43)**2 * 1.e-14 * u.Jy
    wlu = u.Unit(hdr['CUNIT1'])
    wl0 = hdr['CRVAL1']
    dwl = hdr['CDELT1']
    disp = np.arange(wl0, wl0 + (len(flux) - 0.5) * dwl, dwl) * wlu

    spectrum = Spectrum1D(flux=flux, wcs=WCS(hdr))
    tmpfile = tmp_path / 'wcs_tst.fits'
    spectrum.write(tmpfile, hdu=0)

    # Deliberately not using standard filename pattern to test header info.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FITSFixedWarning)
        os.system(f'{compress} {tmpfile}')
        spec = Spectrum1D.read(tmpfile.with_suffix(f'{tmpfile.suffix}{ext[compress]}'))

    assert isinstance(spec, Spectrum1D)
    assert quantity_allclose(spec.spectral_axis, disp)
    assert quantity_allclose(spec.flux, spectrum.flux)

    # Try again without compression suffix:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FITSFixedWarning)
        shutil.move(tmpfile.with_suffix(f'{tmpfile.suffix}{ext[compress]}'), tmpfile)
        spec = Spectrum1D.read(tmpfile)

    assert isinstance(spec, Spectrum1D)
    assert quantity_allclose(spec.spectral_axis, disp)
    assert quantity_allclose(spec.flux, spectrum.flux)


@pytest.mark.remote_data
def test_apstar_loader():
    """Test remote read and automatic recognition of apStar spec from URL.
    """
    apstar_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/"
                  "stars/apo25m/N7789/apStar-r12-2M00005414+5522241.fits")
    spec = Spectrum1D.read(apstar_url)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.flux.unit == 1e-17 * u.erg / (u.s * u.cm**2 * u.AA)
    assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.remote_data
def test_apvisit_loader():
    """Test remote read and automatic recognition of apvisit spec from URL.
    """
    apvisit_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/redux/r12/"
                   "visit/apo25m/N7789/5094/55874/"
                   "apVisit-r12-5094-55874-123.fits")
    spec = Spectrum1D.read(apvisit_url)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.flux.unit == 1e-17 * u.erg / (u.s * u.cm**2 * u.AA)
    assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.remote_data
def test_aspcapstar_loader():
    """Test remote read and automatic recognition of aspcapStar spec from URL.
    """
    aspcap_url = ("https://data.sdss.org/sas/dr16/apogee/spectro/aspcap/r12/"
                  "l33/apo25m/N7789/aspcapStar-r12-2M00005414+5522241.fits")
    spec = Spectrum1D.read(aspcap_url)

    assert isinstance(spec, Spectrum1D)
    assert spec.flux.size > 0
    assert spec.uncertainty.array.min() >= 0.0


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@pytest.mark.remote_data
def test_muscles_loader():
    """Test remote read and automatic recognition of muscles spec from URL.
    """
    url = ("https://archive.stsci.edu/missions/hlsp/muscles/gj1214/"
           "hlsp_muscles_multi_multi_gj1214_broadband_v22_const-res-sed.fits")
    spec = Spectrum1D.read(url)

    assert isinstance(spec, Spectrum1D)
    assert len(spec.flux) == len(spec.spectral_axis) > 50000
    assert spec.uncertainty.array.min() >= 0.0
    assert spec.spectral_axis.unit == u.AA
    assert spec.flux.unit == u.erg / (u.s * u.cm**2 * u.AA)

    # Read HDUList
    with fits.open(url) as hdulist:
        spec = Spectrum1D.read(hdulist, format="MUSCLES SED")
        assert isinstance(spec, Spectrum1D)


@pytest.mark.remote_data
def test_subaru_pfs_loader(tmp_path):
    """Test remote read and automatic recognition of Subaru PFS spec from URL.
    """
    pfs = "pfsObject-00000-0,0-000-00000001-01-0x395428ab.fits"
    url = f"https://github.com/Subaru-PFS/datamodel/raw/master/examples/{pfs}"

    assert subaru_pfs_spec.identify_pfs_spec(url, url)

    # PFS loader parses metadata from filename, cannot read directly from url
    tmpfile = str(tmp_path / pfs)
    with urllib.request.urlopen(url) as response:
        with open(tmpfile, mode='wb') as fout:
            shutil.copyfileobj(response, fout)

    assert subaru_pfs_spec.identify_pfs_spec(pfs, open(tmpfile, mode='rb'))
    spec = Spectrum1D.read(tmpfile, format='Subaru-pfsObject')
    assert isinstance(spec, Spectrum1D)

    spec = Spectrum1D.read(tmpfile)
    assert isinstance(spec, Spectrum1D)
    assert len(spec.flux) == len(spec.spectral_axis) > 10000
    assert spec.spectral_axis.unit == u.nm
    assert spec.flux.unit == u.nJy


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([{'id': '3733958', 'filename': '1D-c0022498-344732.fits'}])
def test_spectrum1d_6dfgs_tabular(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert spec.spectral_axis.unit == u.Unit("Angstrom")
    assert spec.flux.unit == u.Unit("count/s")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="6dFGS-tabular")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.unit == u.Unit("count/s")
    assert spec.flux.size > 0


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([{'id': '3733958', 'filename': 'all-c0022498-344732v_spectrum0.fits'}])
def test_spectrum1d_6dfgs_split_v(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert spec.spectral_axis.unit == u.Unit("Angstrom")
    assert spec.flux.unit == u.Unit("count/Angstrom")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="6dFGS-split")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.unit == u.Unit("count/Angstrom")
    assert spec.flux.size > 0


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([{'id': '3733958', 'filename': 'all-c0022498-344732r_spectrum0.fits'}])
def test_spectrum1d_6dfgs_split_r(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert spec.spectral_axis.unit == u.Unit("Angstrom")
    assert spec.flux.unit == u.Unit("count/Angstrom")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="6dFGS-split")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.unit == u.Unit("count/Angstrom")
    assert spec.flux.size > 0


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([{'id': '3733958', 'filename': 'all-c0022498-344732combined_spectrum0.fits'}])
def test_spectrum1d_6dfgs_split_combined(remote_data_path):
    spec = Spectrum1D.read(remote_data_path)

    assert spec.spectral_axis.unit == u.Unit("Angstrom")
    assert spec.flux.unit == u.Unit("count/Angstrom")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        spec = Spectrum1D.read(hdulist, format="6dFGS-split")
    assert isinstance(spec, Spectrum1D)
    assert spec.flux.unit == u.Unit("count/Angstrom")
    assert spec.flux.size > 0


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([{'id': '3733958', 'filename': 'all-c0022498-344732.fits'}])
def test_spectrum1d_6dfgs_combined(remote_data_path):
    specs = SpectrumList.read(remote_data_path)

    for spec in specs:
        assert spec.spectral_axis.unit == u.Unit("Angstrom")
        assert spec.flux.unit == u.Unit("count/Angstrom")

    assert len(specs) == 3

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        specs = SpectrumList.read(hdulist, format="6dFGS-combined")
        for spec in specs:
            assert isinstance(spec, Spectrum1D)
            assert spec.flux.unit == u.Unit("count/Angstrom")
            assert spec.flux.size > 0
            assert spec.meta["sky"].flux.unit == u.Unit("count/Angstrom")
            assert spec.meta["sky"].flux.size > 0

        assert len(specs) == 3


# Commented out until science only is discussed
# @pytest.mark.remote_data
# def test_2slaq_lrg_loader_science_only():
#     """Test remote read and automatic recognition of 2SLAQ-LRG data from URL.
#     """
#     url = ("https://datacentral.org.au/services/sov/81480/download/"
#            "gama.dr2.spectra.2slaq-lrg.spectrum_1d/J143529.78-004306.4_1.fit/")
#     spec = Spectrum1D.read(url)
#
#     assert spec.spectral_axis.unit == u.AA
#     assert spec.flux.unit == u.count / u.s
#     assert spec.uncertainty is None


@remote_access([{'id': '3970324', 'filename': 'J143529.78-004306.4_1.fit'}])
def test_2slaq_lrg_loader_science_and_sky(remote_data_path):
    """Test remote read and automatic recognition of 2SLAQ-LRG data from URL.
    """
    science, sky = SpectrumList.read(remote_data_path)

    assert science.spectral_axis.unit == u.AA
    assert science.flux.unit == u.count / u.s
    assert science.uncertainty is None

    assert sky.spectral_axis.unit == u.AA
    assert sky.flux.unit == u.count / u.s
    assert sky.uncertainty is None


@remote_access([
    {'id': '3895436', 'filename': '000002.fits'},
    {'id': '3895436', 'filename': '000003.fits'},
    {'id': '3895436', 'filename': '000004.fits'},
])
def test_spectrum_list_2dfgrs_single(remote_data_path):
    specs = SpectrumList.read(remote_data_path)

    assert len(specs) == 1

    for spec in specs:
        assert spec.spectral_axis.unit == u.Unit("Angstrom")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        specs = SpectrumList.read(hdulist, format="2dFGRS")
        for spec in specs:
            assert isinstance(spec, Spectrum1D)
            assert spec.spectral_axis.unit == u.Unit("Angstrom")

        assert len(specs) == 1


@remote_access([{'id': '3895436', 'filename': '000001.fits'}])
def test_spectrum_list_2dfgrs_multiple(remote_data_path):
    specs = SpectrumList.read(remote_data_path)

    assert len(specs) == 2

    for spec in specs:
        assert spec.spectral_axis.unit == u.Unit("Angstrom")

    # Read from HDUList object
    with fits.open(remote_data_path) as hdulist:
        specs = SpectrumList.read(hdulist, format="2dFGRS")
        for spec in specs:
            assert isinstance(spec, Spectrum1D)
            assert spec.spectral_axis.unit == u.Unit("Angstrom")

        assert len(specs) == 2


def test_sdss_wcs_handler():
    sdss_wcs = WCS(naxis=2)
    sdss_wcs.wcs.crval[0] = 3.57880000000000E+00
    sdss_wcs.wcs.cd = [[1.00000000000000E-04, 0], [0, 1]]
    sdss_wcs.wcs.cunit[0] = u.Unit('Angstrom')
    fixed_wcs = _sdss_wcs_to_log_wcs(sdss_wcs)
    dropped_sdss_wcs = sdss_wcs.dropaxis(1)
    dropped_sdss_wcs.wcs.cunit[0] = ''  # Cannot handle units in powers
    sdss_wave = 10 ** dropped_sdss_wcs.pixel_to_world(np.arange(10)) * u.Unit('Angstrom')
    fixed_wave = fixed_wcs.pixel_to_world(np.arange(10))
    assert quantity_allclose(sdss_wave, fixed_wave)


class TestAAOmega2dF:
    @remote_access([{'id': '4460981', 'filename': "OBJ0039red.fits"}])
    def test_with_rwss(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format="Data Central AAOmega",
        )
        assert len(spectra) == 139
        for spec in spectra:
            assert spec.meta.get("label") is not None
            assert spec.meta.get("header") is not None
            assert spec.meta.get("purpose") is not None
            assert spec.meta.get("fibre_index") is not None

    @remote_access([{'id': '4460981', 'filename': "OBJ0032red.fits"}])
    def test_without_rwss(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format="Data Central AAOmega",
        )
        assert len(spectra) == 153
        for spec in spectra:
            assert spec.meta.get("label") is not None
            assert spec.meta.get("header") is not None
            assert spec.meta.get("purpose") is not None
            assert spec.meta.get("fibre_index") is not None

    @remote_access([{'id': '4460981', 'filename': "OBJ0039red.fits"}])
    def test_with_rwss_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 139
        for spec in spectra:
            assert spec.meta.get("label") is not None
            assert spec.meta.get("header") is not None
            assert spec.meta.get("purpose") is not None
            assert spec.meta.get("fibre_index") is not None

    @remote_access([{'id': '4460981', 'filename': "OBJ0032red.fits"}])
    def test_without_rwss_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 153
        for spec in spectra:
            assert spec.meta.get("label") is not None
            assert spec.meta.get("header") is not None
            assert spec.meta.get("purpose") is not None
            assert spec.meta.get("fibre_index") is not None


@remote_access([
    {'id': "4460981", 'filename':"1812260046012353.fits"},  # 4 exts
    {'id': "4460981", 'filename':"1311160005010021.fits"}  # 5 exts
])
def test_galah(remote_data_path):
    spectra = SpectrumList.read(remote_data_path, format="GALAH")
    # Should be main spectra, without sky, and normalised (not in 4 ext)
    nspec = len(spectra)
    if spectra[0].meta["galah_hdu_format"] == 4:
        assert nspec == 2
    elif spectra[0].meta["galah_hdu_format"] == 5:
        assert nspec == 3
    else:
        assert False, "Unknown format"
    # normalised
    if nspec == 3:
        assert spectra[0].flux.unit == u.Unit('')  # dimensionless
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[0].uncertainty is None
        assert spectra[0].meta.get("label") == "normalised spectra"
        assert spectra[0].meta.get("header") is not None

        # drop the normalised spectra, so 4 and 5 should now look the same
        spectra = spectra[1:]

    # main spectra
    assert spectra[0].flux.unit == u.count
    assert spectra[0].spectral_axis.unit == u.Angstrom
    assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
    assert spectra[0].meta.get("label") is not None
    assert spectra[0].meta.get("header") is not None

    # No sky
    assert spectra[1].spectral_axis.unit == u.Angstrom
    assert spectra[1].flux.unit == u.count
    assert isinstance(spectra[1].uncertainty, StdDevUncertainty)
    assert spectra[1].meta.get("label") is not None
    assert spectra[1].meta.get("header") is not None


@remote_access([
    {'id': "4460981", 'filename':"1812260046012353.fits"},  # 4 exts
    {'id': "4460981", 'filename':"1311160005010021.fits"}  # 5 exts
])
def test_galah_guess(remote_data_path):
    spectra = SpectrumList.read(remote_data_path)
    # Should be main spectra, without sky, and normalised (not in 4 ext)
    nspec = len(spectra)
    if spectra[0].meta["galah_hdu_format"] == 4:
        assert nspec == 2
    elif spectra[0].meta["galah_hdu_format"] == 5:
        assert nspec == 3
    else:
        assert False, "Unknown format"

    if remote_data_path.endswith('1311160005010021.fits'):
        pytest.xfail('Format is ambiguous')

    # main spectra
    assert spectra[0].flux.unit == u.count
    assert spectra[0].spectral_axis.unit == u.Angstrom
    assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
    assert spectra[0].meta.get("label") is not None
    assert spectra[0].meta.get("header") is not None

    # normalised
    if nspec == 3:
        assert spectra[1].flux.unit == u.Unit('')  # dimensionless
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert spectra[1].uncertainty is None
        assert spectra[1].meta.get("label") == "normalised spectra"
        assert spectra[1].meta.get("header") is not None

    # No sky
    if nspec == 3:
        assert spectra[2].spectral_axis.unit == u.Angstrom
        assert spectra[2].flux.unit == u.count
        assert isinstance(spectra[2].uncertainty, StdDevUncertainty)
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None
    else:
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert isinstance(spectra[1].uncertainty, StdDevUncertainty)
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None


filename_list = ["bad_file.fits"]


# We cannot use remote_access directly in the MIRI MRS tests, because
# the test functions are called once for every file in the remote access
# list, but we need to feed the entire set to the functions at once. We
# store the file names in a list and use a dummy test method to load the
# list via the remote_access machinery.
#
# MIRI MRS 1D data sets are comprised of 12 files.
# We add one bad file to test the skip/warn on missing file functionality
@remote_access([
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch1-long_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch1-medium_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch1-short_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch2-long_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch2-medium_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch2-short_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch3-long_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch3-medium_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch3-short_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch4-long_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch4-medium_x1d.fits'},
    {'id': '5082863', 'filename': 'combine_dithers_all_exposures_ch4-short_x1d.fits'},
])
def test_loaddata_miri_mrs(remote_data_path):
    filename_list.append(remote_data_path)


# loading from a list of file names
@pytest.mark.filterwarnings(r'ignore:Failed to load bad_file\.fits')
@pytest.mark.remote_data
def test_spectrum_list_names_miri_mrs():

    # Format is explicitly set
    with pytest.raises(FileNotFoundError):
        specs = SpectrumList.read(filename_list, format="JWST x1d MIRI MRS")

    # Skip missing file silently
    specs = SpectrumList.read(filename_list, format="JWST x1d MIRI MRS", missing='silent')

    assert len(specs) == 12
    for spec in specs:
        assert isinstance(spec, Spectrum1D)
        assert spec.spectral_axis.unit == u.micron

    # Warn about missing file
    specs = SpectrumList.read(filename_list, format="JWST x1d MIRI MRS", missing='warn')

    assert len(specs) == 12
    for spec in specs:
        assert isinstance(spec, Spectrum1D)
        assert spec.spectral_axis.unit == u.Unit("um")

    # Auto-detect format
    specs = SpectrumList.read(filename_list[1:])

    assert len(specs) == 12
    for spec in specs:
        assert isinstance(spec, Spectrum1D)
        assert spec.spectral_axis.unit == u.micron


# loading from a directory via glob
@pytest.mark.remote_data
def test_spectrum_list_directory_miri_mrs(tmp_path):

    # copy files to temp dir. We cannot use the directory generated by
    # remote_access, because it may have variable structure from run to
    # run. And also because temp directories created in previous runs
    # may still be hanging around. This precludes the use of commonpath()
    tmp_dir = str(tmp_path)
    for file_path in filename_list[1:]:
        shutil.copy(file_path, tmp_dir)

    specs = SpectrumList.read(tmp_dir)

    assert len(specs) == 12
    for spec in specs:
        assert isinstance(spec, Spectrum1D)
        assert spec.spectral_axis.unit == u.micron


# x1d and c1d assorted files

@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@pytest.mark.filterwarnings('ignore:SRCTYPE is missing or UNKNOWN')
@remote_access([
    {'id': "5394931", 'filename':"jw00623-c1012_t002_miri_p750l_x1d.fits"},                   # pipeline 1.2.3
    {'id': "5394931", 'filename':"jw00787-o014_s00002_niriss_f150w-gr150c-gr150r_c1d.fits"},  # pipeline 1.2.3
    {'id': "5394931", 'filename':"jw00623-o057_t008_miri_ch1-long_x1d.fits"},                 # pipeline 1.3.1
    {'id': "5394931", 'filename':"jw00626-o064_t007_nirspec_g235h-f170lp_x1d.fits"},          # pipeline 1.3.1
])
def test_jwst_x1d_c1d(remote_data_path):

    data = Spectrum1D.read(remote_data_path)

    assert isinstance(data, Spectrum1D)
    assert data.shape in [(388,), (5,), (1091,), (3843,)]
    assert data.unit == u.Jy
    assert data.spectral_axis.unit == u.um


# utility functions to be used with list comprehension in SpectrumList checking
def assert_multi_isinstance(a, b):
    assert isinstance(a, b)


def assert_multi_equals(a, b):
    assert a == b


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([
    {'id': "5394931", 'filename':"jw00624-o027_s00001_nircam_f356w-grismr_x1d.fits"},  # pipeline 1.2.3
])
def test_jwst_nircam_x1d_multi_v1_2_3(remote_data_path):

    data = SpectrumList.read(remote_data_path)

    assert isinstance(data, SpectrumList)
    assert len(data) == 3
    [assert_multi_isinstance(d, Spectrum1D) for d in data]
    [assert_multi_equals(d.shape, r) for d,r in zip(data, [(459,), (336,), (962,)])]
    [assert_multi_equals(d.unit, u.Jy) for d in data]
    [assert_multi_equals(d.spectral_axis.unit, u.um) for d in data]


@pytest.mark.filterwarnings(r'ignore:.*did not parse as fits unit')
@remote_access([
    {'id': "5394931", 'filename':"jw00660-o016_s00002_nircam_f444w-grismr_x1d.fits"},  # pipeline 1.3.1
])
def test_jwst_nircam_x1d_multi_v1_3_1(remote_data_path):

    data = SpectrumList.read(remote_data_path)

    assert isinstance(data, SpectrumList)
    assert len(data) == 4
    [assert_multi_isinstance(d, Spectrum1D) for d in data]
    [assert_multi_equals(d.shape, r) for d,r in zip(data, [(1166,), (786,), (1157,), (795,)])]
    [assert_multi_equals(d.unit, u.Jy) for d in data]
    [assert_multi_equals(d.spectral_axis.unit, u.um) for d in data]


@remote_access([
    {'id': "5394931", 'filename':"jw00776-o003_s00083_nircam_f322w2-grismr_c1d.fits"},  # pipeline 1.2.3
])
def test_jwst_nircam_c1d_v1_2_3(remote_data_path):

    data = SpectrumList.read(remote_data_path)

    assert isinstance(data, SpectrumList)
    assert len(data) == 2
    [assert_multi_isinstance(d, Spectrum1D) for d in data]
    [assert_multi_equals(d.shape, r) for d,r in zip(data, [(133,), (1139,)])]
    [assert_multi_equals(d.unit, u.MJy/u.sr) for d in data]
    [assert_multi_equals(d.spectral_axis.unit, u.um) for d in data]


@remote_access([
    {'id': "5394931", 'filename':"jw00625-o018_s00001_niriss_f090w-gr150c_c1d.fits"},  # pipeline 1.2.3
])
def test_jwst_niriss_c1d_v1_2_3(remote_data_path):

    data = SpectrumList.read(remote_data_path)

    assert isinstance(data, SpectrumList)
    assert len(data) == 2
    [assert_multi_isinstance(d, Spectrum1D) for d in data]
    [assert_multi_equals(d.shape, r) for d,r in zip(data, [(56,), (107,)])]
    [assert_multi_equals(d.unit, u.Jy) for d in data]
    [assert_multi_equals(d.spectral_axis.unit, u.um) for d in data]


class TestSAMI:
    @remote_access([
        {'id': "10802828", 'filename':"24433_A_adaptive_blue.fits.gz"},
        {'id': "10802828", 'filename':"24433_A_adaptive_red.fits.gz"},
        {'id': "10802828", 'filename':"24433_A_annular_blue.fits.gz"},
        {'id': "10802828", 'filename':"24433_A_annular_red.fits.gz"},
        {'id': "10802828", 'filename':"24433_A_cube_blue.fits.gz"},
        {'id': "10802828", 'filename':"24433_A_cube_red.fits.gz"},
        {'id': "10802828", 'filename':"24433_adaptive_blue.fits.gz"},
        {'id': "10802828", 'filename':"24433_adaptive_red.fits.gz"},
        {'id': "10802828", 'filename':"24433_spectrum_1-4-arcsec_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_1-4-arcsec_red.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_2-arcsec_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_2-arcsec_red.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_3-arcsec_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_3-arcsec_red.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_3-kpc_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_3-kpc_red.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_4-arcsec_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_4-arcsec_red.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_re_blue.fits"},
        {'id': "10802828", 'filename':"24433_spectrum_re_red.fits"},
    ])
    def test_sami_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 1

        spec = spectra[0]
        assert isinstance(spec.uncertainty, VarianceUncertainty)
        assert spec.flux.unit == u.Unit('10**(-16) erg/s/cm**2/angstrom/pixel')

        if len(spec.flux.shape) == 3:
            # This is a cube
            assert spec.flux.shape == (50, 50, 2048)
            assert "sami_QC_table" in spec.meta
            assert "sami_dust_vector_weights" in spec.meta

        else:
            # This is a 1D spectrum
            assert spec.flux.shape == (2048,)
            assert "sami_aperture_spectra_mask" in spec.meta



"""Helpers for testing specutils objects in ASDF files.
These are similar to those in ``asdf_astropy.testing.helpers``.
"""
from asdf_astropy.testing.helpers import assert_spectral_coord_equal
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose, assert_array_equal

__all__ = ["assert_spectral_axis_equal", "assert_spectrum1d_equal", "assert_spectrumlist_equal"]


def assert_spectral_axis_equal(a, b):
    """Equality test for use in ASDF unit tests for SpectralAxis."""
    __tracebackhide__ = True

    assert_spectral_coord_equal(a, b)


def assert_spectrum1d_equal(a, b):
    """Equality test for use in ASDF unit tests for Spectrum1D."""
    __tracebackhide__ = True

    assert_quantity_allclose(a.flux, b.flux)
    assert_spectral_axis_equal(a.spectral_axis, b.spectral_axis)

    if a.uncertainty is None:
        assert b.uncertainty is None
    else:
        assert a.uncertainty.uncertainty_type == b.uncertainty.uncertainty_type
        assert_allclose(a.uncertainty.array, b.uncertainty.array)

    if a.mask is None:
        assert b.mask is None
    else:
        assert_array_equal(a.mask, b.mask)


def assert_spectrumlist_equal(a, b):
    """Equality test for use in ASDF unit tests for SpectrumList."""
    __tracebackhide__ = True

    assert len(a) == len(b)
    for x, y in zip(a, b):
        assert_spectrum1d_equal(x, y)



# ~/.specutils/my_custom_loader.py
#
#  Load a FITS cube , extract the spectrum at the reference pixel, but this
#  can be optionally overriden.
#
#  21-apr-2016  Peter Teuben    hackday at "SPECTROSCOPY TOOLS IN PYTHON WORKSHOP" STSCI

import numpy as np
from astropy.units import Unit
from astropy.wcs import WCS

from ...spectra import Spectrum1D
from ..parsing_utils import read_fileobj_or_hdulist


# Define an optional identifier. If made specific enough, this circumvents the
# need to add `format="my-format"` in the `Spectrum1D.read` call.
def identify_generic_fits(origin, *args, **kwargs):
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (hdulist[0].header['NAXIS'] == 3)


# not yet ready because it's not generic enough and does not use column_mapping
# @data_loader("Cube", identifier=identify_generic_fits, extensions=['fits'])
def generic_fits(file_obj, **kwargs):

    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        data3  = hdulist[0].data
        wcs    = WCS(header)
        shape  = data3.shape

        # take the reference pixel if the pos= was not supplied by the reader
        if 'pos' in kwargs:
            ix = kwargs['pos'][0]
            iy = kwargs['pos'][1]
        else:
            ix = int(wcs.wcs.crpix[0])
            iy = int(wcs.wcs.crpix[1])

        # grab a spectrum from the cube
        if len(shape) == 3:
            data = data3[:,iy,ix]
        elif len(shape) == 4:
            data = data3[:,:,iy,ix].squeeze()
            # make sure this is a 1D array
            # if len(data.shape) != 1:
            #    raise Exception("not a true cube")
        else:
            raise ValueError(f"Unexpected shape {shape}.")

        # store some meta data
        meta = {'header': header}
        meta['xpos'] = ix
        meta['ypos'] = iy

        # attach units (get it from header['BUNIT'] - what about 'JY/BEAM '
        #  NOTE:  astropy doesn't support beam, but see comments in radio_beam
        data = data * Unit("Jy")

        # now figure out the frequency axis....
        sp_axis = 3
        naxis3 = header['NAXIS%d' % sp_axis]
        cunit3 = wcs.wcs.cunit[sp_axis-1]
        crval3 = wcs.wcs.crval[sp_axis-1]
        cdelt3 = wcs.wcs.cdelt[sp_axis-1]
        crpix3 = wcs.wcs.crpix[sp_axis-1]

        freqs = np.arange(naxis3) + 1
        freqs = (freqs - crpix3) * cdelt3 + crval3

        freqs = freqs * cunit3

        # should wcs be transformed to a 1D case ?

    return Spectrum1D(flux=data, wcs=wcs, meta=meta, spectral_axis=freqs)
    # return Spectrum1D(flux=data, wcs=wcs, meta=meta)     # this does not work yet



import astropy.units as u
import numpy as np
from astropy import constants as const
from astropy.nddata import StdDevUncertainty
from astropy.units import Quantity
from scipy.signal.windows import tukey
from scipy.signal import correlate

from ..manipulation import LinearInterpolatedResampler
from .. import Spectrum1D

__all__ = ['template_correlate', 'template_logwl_resample']

_KMS = u.Unit('km/s')  # for use below without having to create a composite unit


def template_correlate(observed_spectrum, template_spectrum, lag_units=_KMS,
                       apodization_window=0.5, resample=True, method="direct"):
    """
    Compute cross-correlation of the observed and template spectra.


    After re-sampling into log-wavelength, both observed and template
    spectra are apodized by a Tukey window in order to minimize edge
    and consequent non-periodicity effects and thus decrease
    high-frequency power in the correlation function. To turn off the
    apodization, use alpha=0.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum, which will be correlated with
        the observed spectrum.
    lag_units: `~astropy.units.Unit`
        Must be a unit with velocity physical type for lags in velocity. To
        output the lags in redshift, use ``u.dimensionless_unscaled``.
    apodization_window: float, callable, or None
        If a callable, will be treated as a window function for apodization of
        the cross-correlation (should behave like a `~scipy.signal.windows`
        window function, with ``sym=True``). If a float, will be treated as the
        ``alpha`` parameter for a Tukey window (`~scipy.signal.windows.tukey`),
        in units of pixels. If None, no apodization will be performed
    resample: bool or dict
        If True or a dictionary, resamples the spectrum and template following
        the process in `template_logwl_resample`. If a dictionary, it will be
        used as the keywords for `template_logwl_resample`.  For example,
        ``resample=dict(delta_log_wavelength=.1)`` would be the same as calling
        ``template_logwl_resample(spectrum, template, delta_log_wavelength=.1)``.
        If False, *no* resampling is performed (and the user is responsible for
        a sensible resampling).
    method: str
        If you choose "FFT", the correlation will be done through the use
        of convolution and will be calculated faster (for small spectral
        resolutions it is often correct), otherwise the correlation is determined
        directly from sums (the "direct" method in `~scipy.signal.correlate`).

    Returns
    -------
    (`~astropy.units.Quantity`, `~astropy.units.Quantity`)
        Arrays with correlation values and lags in km/s
    """

    # resample if the user requested to log wavelength
    if resample:
        if resample is True:
            resample_kwargs = dict()  # use defaults
        else:
            resample_kwargs = resample
        log_spectrum, log_template = template_logwl_resample(observed_spectrum,
                                                             template_spectrum,
                                                             **resample_kwargs)
    else:
        log_spectrum = observed_spectrum
        log_template = template_spectrum

    # apodize (might be a no-op if apodization_window is None)
    observed_log_spectrum, template_log_spectrum = _apodize(log_spectrum,
                                                            log_template,
                                                            apodization_window)
    # Normalize template
    normalization = _normalize(observed_log_spectrum, template_log_spectrum)

    # Not sure if we need to actually normalize the template. Depending
    # on the specific data uncertainty, the normalization factor
    # may turn out negative. That causes a flip of the correlation function,
    # in which the maximum (correlation peak) is no longer meaningful.
    if normalization < 0.:
        normalization = 1.

    # Correlate
    if method.lower() == "fft":
        corr = correlate(observed_log_spectrum.flux.value,
                         (template_log_spectrum.flux.value * normalization),
                         method="fft")
    else:
        corr = correlate(observed_log_spectrum.flux.value,
                         (template_log_spectrum.flux.value * normalization),
                         method="direct")

    # Compute lag
    # wave_l is the wavelength array equally spaced in log space.
    wave_l = observed_log_spectrum.spectral_axis.value
    delta_log_wave = np.log10(wave_l[1]) - np.log10(wave_l[0])
    deltas = (np.array(range(len(corr))) - len(corr)/2 + 0.5) * delta_log_wave
    lags = np.power(10., deltas) - 1.

    if u.dimensionless_unscaled.is_equivalent(lag_units):
        lags = Quantity(lags, u.dimensionless_unscaled)
    elif _KMS.is_equivalent(lag_units):
        lags = lags * const.c.to(lag_units)
    else:
        raise u.UnitsError('lag_units must be either velocity or dimensionless')

    return corr * u.dimensionless_unscaled, lags


def _apodize(spectrum, template, apodization_window):
    # Apodization. Must be performed after resampling.
    if apodization_window is None:
        clean_spectrum = spectrum
        clean_template = template
    else:
        if callable(apodization_window):
            window = apodization_window
        else:
            def window(wlen):
                return tukey(wlen, alpha=apodization_window)
        clean_spectrum = spectrum * window(len(spectrum.spectral_axis))
        clean_template = template * window(len(template.spectral_axis))

    return clean_spectrum, clean_template


def template_logwl_resample(spectrum, template, wblue=None, wred=None,
                            delta_log_wavelength=None,
                            resampler=LinearInterpolatedResampler()):
    """
    Resample a spectrum and template onto a common log-spaced spectral grid.

    If wavelength limits are not provided, the function will use
    the limits of the merged (observed+template) wavelength scale
    for building the log-wavelength scale.

    For the wavelength step, the function uses either the smallest wavelength
    interval found in the *observed* spectrum, or takes it from the
    ``delta_log_wavelength`` parameter.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum.
    wblue, wred: float
        Wavelength limits to include in the correlation.
    delta_log_wavelength: float
        Log-wavelength step to use to build the log-wavelength
        scale. If None, use limits defined as explained above.
    resampler
        A specutils resampler to use to actually do the resampling.  Defaults to
        using a `~specutils.manipulation.LinearInterpolatedResampler`.

    Returns
    -------
    resampled_observed : :class:`~specutils.Spectrum1D`
        The observed spectrum resampled to a common spectral_axis.
    resampled_template: :class:`~specutils.Spectrum1D`
        The template spectrum resampled to a common spectral_axis.
    """

    # Build an equally-spaced log-wavelength array based on
    # the input and template spectrum's limit wavelengths and
    # smallest sampling interval. Consider only the observed spectrum's
    # sampling, since it's the one that counts for the final accuracy
    # of the correlation. Alternatively, use the wred and wblue limits,
    # and delta log wave provided by the user.
    #
    # We work with separate float and units entities instead of Quantity
    # instances, due to the profusion of log10 and power function calls
    # (they only work on floats)
    if wblue:
        w0 = np.log10(wblue)
    else:
        ws0 = np.log10(spectrum.spectral_axis[0].value)
        wt0 = np.log10(template.spectral_axis[0].value)
        w0 = min(ws0, wt0)

    if wred:
        w1 = np.log10(wred)
    else:
        ws1 = np.log10(spectrum.spectral_axis[-1].value)
        wt1 = np.log10(template.spectral_axis[-1].value)
        w1 = max(ws1, wt1)

    if delta_log_wavelength is None:
        ds = np.log10(spectrum.spectral_axis.value[1:]) - np.log10(spectrum.spectral_axis.value[:-1])
        dw = ds[np.argmin(ds)]
    else:
        dw = delta_log_wavelength

    nsamples = int((w1 - w0) / dw)

    log_wave_array = np.ones(nsamples) * w0
    for i in range(nsamples):
        log_wave_array[i] += dw * i

    # Build the corresponding wavelength array
    wave_array = np.power(10., log_wave_array) * spectrum.spectral_axis.unit

    # Resample spectrum and template into wavelength array so built
    resampled_spectrum = resampler(spectrum, wave_array)
    resampled_template = resampler(template, wave_array)

    # Resampler leaves Nans on flux bins that aren't touched by it.
    # We replace with zeros. This has the net effect of zero-padding
    # the spectrum and/or template so they exactly match each other,
    # wavelengthwise.
    clean_spectrum_flux = np.nan_to_num(resampled_spectrum.flux.value) * resampled_spectrum.flux.unit
    clean_template_flux = np.nan_to_num(resampled_template.flux.value) * resampled_template.flux.unit

    clean_spectrum = Spectrum1D(spectral_axis=resampled_spectrum.spectral_axis,
                        flux=clean_spectrum_flux,
                        uncertainty=resampled_spectrum.uncertainty,
                        velocity_convention='optical',
                        rest_value=spectrum.rest_value)
    clean_template = Spectrum1D(spectral_axis=resampled_template.spectral_axis,
                        flux=clean_template_flux,
                        uncertainty=resampled_template.uncertainty,
                        velocity_convention='optical',
                        rest_value=template.rest_value)

    return clean_spectrum, clean_template


def _normalize(observed_spectrum, template_spectrum):
    """
    Calculate a scale factor to be applied to the template spectrum so the
    total flux in both spectra will be the same.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum, which needs to be normalized in order to be
        compared with the observed spectrum.

    Returns
    -------
    `float`
        A float which will normalize the template spectrum's flux so that it
        can be compared to the observed spectrum.
    """
    if hasattr(observed_spectrum, "uncertainty") and observed_spectrum.uncertainty is not None:
        unc = observed_spectrum.uncertainty.represent_as(StdDevUncertainty)
        num = np.nansum((observed_spectrum.flux*template_spectrum.flux)/(unc.array**2))
        denom = np.nansum((template_spectrum.flux/unc.array)**2)
    else:
        num = np.nansum(observed_spectrum.flux*template_spectrum.flux)
        denom = np.nansum(template_spectrum.flux**2)

    return num/denom
